"""
    DecompEditorFrame.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import copy
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.gridspec    import GridSpec
import matplotlib.patches   as mpl_patches      # 'as patches' does not work properly
from matplotlib import colors
import matplotlib
import seaborn as sns
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.OurMatplotlib import get_color, get_hex_color
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from .SuperEditorFrame import SuperEditorFrame
from molass_legacy.DataStructure.PeakInfo import PeakInfo
from molass_legacy.DataStructure.RangeInfo import DecompEditorInfo, shift_editor_ranges
from molass_legacy.DataStructure.AnalysisRangeInfo import AnalysisRangeInfo, convert_to_paired_ranges
from molass_legacy.SerialAnalyzer.PairedRangeLogger import log_paired_ranges
from molass_legacy.Decomposer.DecompUtils import (
                                    make_range_info_impl,
                                    decompose_elution_better,
                                    )
from molass_legacy.Decomposer.UnifiedDecompResult import UnifiedDecompResult
from .DecompSpecPanel import SpecPanel, RANGE_FRAME_HEIGHT
from molass_legacy.SerialAnalyzer.DataUtils import compact_path_name

MIN_NUM_PANEL_SPACES    = 3
CANVAS_HIGHT    = 6
DECOMP_EDITOR_LOG_HEADER  = "--- decomp editor log begin ---"
DECOMP_EDITOR_LOG_TRAILER = "--- decomp editor log end ---"
RESTORE_HINTS_DICT = False      # TODO

toggle_button_texts = [ "Show Xray decomposition", "Show UV decomposition" ]

def format_coord(x, y):
    return 'x=%.4g    y=%.4g' % (x, y)

class ModelDecompInfo:
    def __init__(self, decomp_ret):
        self.model_name     = decomp_ret.model_name
        self.opt_recs       = decomp_ret.opt_recs
        self.opt_recs_uv    = decomp_ret.opt_recs_uv
        self.num_eltns      = len(self.opt_recs)

class DecompEditorFrame(SuperEditorFrame):
    def __init__(self, parent, dialog, sd, mapper, corbase_info, model, decomp_result=None, global_flag=False):
        self.dialog = dialog
        self.logger = dialog.logger
        Tk.Frame.__init__(self, parent)

        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'

        self.mapper = mapper
        self.xr_j0 = sd.xr_j0
        self.corbase_info = corbase_info
        self.model = model
        name = model.get_name()
        self.model_name = name[0:3]     # task: add a method for this name

        memorized_model = get_setting('editor_model')
        range_info = get_setting('decomp_editor_info')  # temporary fix
        self.is_memorized = memorized_model is not None and memorized_model == self.model_name and range_info is not None

        self.params_controllable = True
        self.elution_fig_type = 0   # 0 : Xray, 1:UV
        self.popup_menu = None
        self.ex_solver = None

        self.recompute_decomposition(decomp_result=decomp_result, redraw=False)
        self.decomp_result_init = self.decomp_result

        self.toggle = True
        gf = 'Global ' if global_flag else ''
        in_folder_name = compact_path_name(get_setting('in_folder'))
        title = "%sDecomposition Specification for %s using %s" % (gf, in_folder_name, self.model_name)
        SuperEditorFrame.__init__(self, parent, title)
        if global_flag:
            self.toggle_btn.config(state=Tk.DISABLED)

    def is_delayed(self):
        return self.model.is_delayed()

    def debug_plot(self):
        from importlib import reload
        from molass_legacy.DataStructure.AnalysisRangeInfo import AnalysisRangeInfo
        import molass_legacy.Decomposer.DecompEditorDebug
        reload(molass_legacy.Decomposer.DecompEditorDebug)
        from molass_legacy.Decomposer.DecompEditorDebug import debug_plot_impl
        ranges = AnalysisRangeInfo(self.make_range_info()).get_ranges()
        debug_plot_impl(self, self.dialog)

    def get_figsize(self):
        fig_height = min( CANVAS_HIGHT, self.num_eltns*2 )
        figsize = (3, fig_height)
        print("figsize=", figsize)
        return figsize

    def update_error_label(self):
        text='fit_error=%.3g' % self.fit_error
        self.error_label.config(text=text)

    def draw1(self, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.RangeEditors.DecompPlotUtils
            reload(molass_legacy.RangeEditors.DecompPlotUtils)

        from .DecompPlotUtils import draw1_impl
        draw1_impl(self)

    def write_residual_amount_to_testerlog(self, y, resid_y):
        from molass_legacy.Test.TesterLogger import write_to_tester_log

        max_y = np.percentile(y, 95)
        normalized_resid_amount = np.average(np.abs(resid_y)) / len(resid_y) / max_y
        type_ = 'Xray' if self.elution_fig_type == 0 else 'UV'

        write_to_tester_log( "normalized_resid_amount=%g with %s\n" % (normalized_resid_amount, type_))

    def draw2(self, peak_no=None, debug=False):
        self.fig2.clear()
        figsize = self.get_figsize()
        self.fig2.set_size_inches(*figsize, forward=True)   # forward=True is required when the canvas is resized
                                                            # see also self.refresh_figs below

        fx = self.fx
        x  = self.x
        if self.elution_fig_type == 0:
            y   = self.y
            color = 'orange'
            opt_recs = self.opt_recs
        else:
            y   = self.uv_y
            linestyle = ':'
            color = 'blue'
            opt_recs = self.opt_recs_uv

        num_eltns = self.num_eltns
        gs = GridSpec( num_eltns, 1 )
        self.axes = [ self.fig2.add_subplot( gs[i] ) for i in range(num_eltns) ]

        self.fig2_range_parts_list = []
        for k, ax in enumerate(self.axes):
            if debug:
                print([k], peak_no)
            ax.cla()
            # ax.set_axis_off()
            ax.axes.get_xaxis().set_ticks([])
            ax.axes.get_yaxis().set_ticks([])
            ax.plot(x, y, color=color)
            ee_select = self.select_matrix[k]
            for j, e in enumerate(ee_select):
                if debug:
                    print("j, e=", j, e)
                if e == 1:
                    func = opt_recs[j][1]
                    ax.plot(x, func(fx), ':', color=get_color(j), linewidth=3)

            range_parts = []

            if  ( len(self.specpanel_list) == 0 and not self.ignorable_flags[k]
                or len(self.specpanel_list) > 0 and self.specpanel_list[k].ignore.get() == 0 ):
                self.add_range_patchs(ax, range_parts, k=k)

            self.fig2_range_parts_list.append(range_parts)

    def add_range_patchs(self, ax, range_parts, k=None):
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        range_parts.append([ymin, ymax])

        if k is None:
            editor_ranges_ = self.editor_ranges
        else:
            editor_ranges_ = [self.editor_ranges[k]]

        for j, range_list in enumerate(editor_ranges_):
            if k is None and self.ignorable_flags[j]:
                range_parts.append([j])
                continue

            for f, t in range_list:
                p = mpl_patches.Rectangle(
                        (f, ymin),  # (x,y)
                        t - f,   # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.2,
                    )
                ax.add_patch( p )
                lines = []
                for x in [f, t]:
                    line, = ax.plot( [x, x], [ymin, ymax], ':', color='gray', alpha=0.5 )
                    lines.append(line)
                range_parts.append([j, p, lines])

    def refresh_figs(self, devel=False):
        if devel:
            from importlib import reload
            import molass_legacy.KekLib.TkMplUtils
            reload(molass_legacy.KekLib.TkMplUtils)
        from molass_legacy.KekLib.TkMplUtils import adjust_the_tkframe_size
        self.draw1()
        self.draw2()
        self.canvas_draw1()
        self.canvas_draw2()
        # self.update()
        adjust_the_tkframe_size(self.fig2, self.mpl_canvas2_widget, self.pcframe)    # call this after fig2 is drawn

    def update1(self):
        print('update1 start')
        try:
            self.update1_impl(self.fig1_range_parts)
            self.update1_impl(self.fig1_range_parts_resid)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)
        print('update1 end')

    def update1_impl(self, range_parts):
        print('update1_impl start')
        ymin, ymax = range_parts[0]
        for k, parts in enumerate(range_parts[1:]):
            if parts is None:
                continue

            j, patch, lines = parts
            range_list = self.editor_ranges[j]
            print([k, j], 'range_list=', range_list)
            for f, t in range_list:
                x, y = patch.xy
                patch.set_xy((f, y))
                patch.set_width(t - f)
                for n, x_ in enumerate([f, t]):
                    line = lines[n]
                    line.set_data([x_, x_], [ymin, ymax])

    def update2(self, peak_no):
        print('update2: peak_no=', peak_no)
        pk = peak_no - 1
        range_parts = self.fig2_range_parts_list[pk]

        ymin, ymax = range_parts[0]
        for k, parts in enumerate(range_parts[1:]):
            if parts is None:
                continue

            j, patch, lines = parts
            range_list = self.editor_ranges[j]
            for f, t in range_list:
                x, y = patch.xy
                patch.set_xy((f, y))
                patch.set_width(t - f)

    def update_figs(self, peak_no):
        self.update1()
        self.update2(peak_no)
        self.canvas_draw1()
        self.canvas_draw2()

    def add_widgets(self):
        num_eltns = self.num_eltns

        canvas_width = int(self.mpl_canvas2_widget.cget('width')) + self.pcframe_padx*2
        self.col_widths  = [canvas_width, 60, 120, 80, 60, 160, 60, 60, 160]
        # self.col_widths[2] = 120 required for SUB_TRN1 with EMG
        label_texts = [ 'Figure',
                        'Peak No',
                        'Element Curve(s)',
                        'Number of\nRanges',
                        'Range Id',
                        'Range(s)\nFrom       To',
                        'Ignore',
                        ]
        if self.params_controllable:
            label_texts.append( 'τ Value' )
            label_texts.append( 'τ Constraints' )

        self.col_title_frames = []
        for k, text in enumerate(label_texts):
            frame = Tk.Frame(self.ptframe, width=self.col_widths[k], height=40)
            frame.grid(row=0, column=k, sticky=Tk.S)
            self.col_title_frames.append(frame)
            # https://stackoverflow.com/questions/16363292/label-width-in-tkinter
            frame.pack_propagate(0)
            label = Tk.Label(frame, text=text)
            label.pack(side=Tk.BOTTOM)

        if self.params_controllable:
            self.col_title_frames[-2].grid_forget()
            self.col_title_frames[-1].grid_forget()

        for k in range(max(MIN_NUM_PANEL_SPACES, num_eltns)):
            self.ppframe.grid_rowconfigure(k, weight=1)

        if self.is_memorized and RESTORE_HINTS_DICT:
            hints_dict = get_setting('tau_hints_dict')
            self.logger.info('restored hints_dict will be used: ' + str(hints_dict))
        else:
            hints_dict = None
        self.refresh_spec_panels(hints_dict=hints_dict)

        for k in range( max(0, MIN_NUM_PANEL_SPACES - num_eltns) ):
            space = Tk.Frame(self.ppframe, height=RANGE_FRAME_HEIGHT)
            space.grid(row=num_eltns+k, column=0)

    def make_range_info(self, concentration_datatype=2, debug=False):
        """
            called at PreviewButtonFrame.get_preview_data()
        """
        shifted_ranges = shift_editor_ranges(-self.xr_j0, self.editor_ranges)
        self.control_info.editor_ranges = shifted_ranges

        if concentration_datatype == 0:       # Xray elution model
            from importlib import reload
            import molass_legacy.Decomposer.OptRecsUtils
            reload(molass_legacy.Decomposer.OptRecsUtils)
            from molass_legacy.Decomposer.OptRecsUtils import eoii_correct_opt_recs
            opt_recs_ = copy.deepcopy(self.opt_recs)
            eoii_correct_opt_recs(self, opt_recs_)

        elif concentration_datatype == 1:     # Xray measured data
            opt_recs_ = self.opt_recs   # will not be used, currenctly
        else:
            opt_recs_ = self.opt_recs_uv

        set_setting('concentration_datatype', concentration_datatype)
        return make_range_info_impl(opt_recs_, self.control_info, self.specpanel_list, logger=self.logger, debug=debug, parent=self)

    def make_restorable_ranges(self):
        paired_ranges = []
        for panel in self.specpanel_list:
            paired_ranges.append(panel.get_paired_range())
        return paired_ranges

    def apply( self ):  # overrides parent class method
        set_setting( 'use_elution_models', 1 )

        analysis_range_info = AnalysisRangeInfo( self.make_range_info(), editor='DecompEditor' )
        print('analysis_range_info=', analysis_range_info)
        set_setting( 'analysis_range_info', analysis_range_info )
        set_setting( 'range_type', 4 )
        paired_ranges = self.make_restorable_ranges()
        set_setting( 'decomp_editor_info', DecompEditorInfo(paired_ranges, self.ignorable_flags))
        set_setting( 'editor_model', self.model_name )

        if RESTORE_HINTS_DICT:
            hints_dict = self.get_param_hints_dict()
            set_setting( 'tau_hints_dict', hints_dict )

        ret = convert_to_paired_ranges(analysis_range_info.get_ranges())
        paired_ranges = ret[0]
        self.log_info_for_reset()
        log_paired_ranges(self.logger, paired_ranges)

    def set_data_label(self):
        data_type = "Xray" if self.elution_fig_type == 0 else "UV"
        self.data_label.config( text=data_type + " elution decomposition" )

    def set_toggle_text(self):
        self.toggle_btn.config( text=toggle_button_texts[1-self.elution_fig_type] )

    def toggle_show( self ):
        self.elution_fig_type = 1 - self.elution_fig_type
        self.refresh_frame()
        self.dialog.update_global_btn_state(self.elution_fig_type)

    def refresh_frame(self, elution_fig_type=None):
        if elution_fig_type is not None:
            self.elution_fig_type = elution_fig_type
        self.set_data_label()
        self.set_toggle_text()
        self.refresh_figs()
        self.update_error_label()

    def reset_to_defaults( self ):
        pass

    def get_fit_error( self ):
        return self.fit_error

    def toggle_params_constraints( self ):
        btn_text = self.param_btn.cget('text')
        is_show_button = btn_text.find("Show") >= 0
        if is_show_button:
            k = len(self.col_title_frames) - 1
            self.col_title_frames[-2].grid(row=0, column=k-1, sticky=Tk.S)
            self.col_title_frames[-1].grid(row=0, column=k, sticky=Tk.S)
        else:
            self.col_title_frames[-2].grid_forget()
            self.col_title_frames[-1].grid_forget()

        for specpanel in self.specpanel_list:
            if is_show_button:
                specpanel.constraints_restore()
            else:
                specpanel.constraints_forget()

        if is_show_button:
            text_ = btn_text.replace("Show", "Hide")
            self.recompute_btn.grid(row=0, column=3, padx=5)
        else:
            text_ = btn_text.replace("Hide", "Show")
            self.recompute_btn.grid_forget()

        self.param_btn.config(text=text_)

        self.update()   # necessary to make effective the following xview_moveto

        if is_show_button:
            if self.dialog.scrollable:
                self.dialog.scrolled_frame.canvas.xview_moveto(1)

    def recompute_decomposition(self, decomp_result=None, redraw=True , debug=False):
        if debug:
            print('recompute_decomposition')
            
        hints_dict = None   # task: remove this hints_dict

        init_mode = decomp_result is None

        if init_mode:
            if self.model.is_traditional():
                decomp_result = decompose_elution_better(self.corbase_info, self.mapper, self.model,
                                    hints_dict=hints_dict, logger=self.logger)
            else:
                if self.model_name == 'STC':
                    from importlib import reload
                    import molass_legacy.Selective.StochasticAdapter
                    reload(molass_legacy.Selective.StochasticAdapter)
                    from molass_legacy.Selective.StochasticAdapter import convert_to_stochastic_decomposition
                    decomp_result = convert_to_stochastic_decomposition(self.dialog)
                else:
                    # i.e., EDM 
                    from importlib import reload
                    import molass_legacy.Selective.AdvancedAdapter
                    reload(molass_legacy.Selective.AdvancedAdapter)
                    from molass_legacy.Selective.AdvancedAdapter import adapted_decompose_elution_simply
                    decomp_result = adapted_decompose_elution_simply(self)

        self.decomp_result = decomp_result
        if init_mode:
            self.decomp_result_init = decomp_result

        # TODO: simple_args=simple_args, use_emga=use_emga, dual_opt=dual_opt

        # task: remove this branch
        if self.model_name == 'STC':
            self.fx = decomp_result.x
            self.x = decomp_result.x
        else:
            self.fx = decomp_result.x
            self.x = self.xr_j0 + decomp_result.x

        self.y = decomp_result.y
        self.uv_y = decomp_result.uv_y
        self.opt_recs = decomp_result.opt_recs
        self.opt_recs_uv = decomp_result.opt_recs_uv
        self.num_eltns = len(decomp_result.opt_recs)

        # be aware that we are getting info from Xray fit recs
        # debug = self.model_name == 'STC'
        self.control_info = decomp_result.get_range_edit_info(logger=self.logger, debug=False)

        self.select_matrix  = self.control_info.select_matrix
        self.top_x_list     = self.control_info.top_x_list

        if self.is_memorized and False:
            # task: reconsider this case
            range_info = get_setting('decomp_editor_info')
            range_info.update(self.control_info)
            assert range_info is not None

            self.ignorable_flags = range_info.get_ignorable_flags()
            self.logger.info('ignorable_flags have been restored as ' + str(self.ignorable_flags))
            ranges = range_info.get_ranges()
            self.editor_ranges  = shift_editor_ranges(self.xr_j0, ranges)
            self.logger.info('editor_ranges have been restored as ' + str(self.editor_ranges))

            self.control_info.editor_ranges = ranges
            """
            the above is required since the following are done to or from self.control_info
                1) SpecPanel.range_list( == control_info.editor_ranges[k]) update
                2) self.make_range_info(), i.e., make_range_info_impl(...)
            """
        else:
            self.ignorable_flags = decomp_result.identify_ignorable_elements()
            self.logger.info('ignorable_flags have been initialized to %s', self.ignorable_flags)
            self.editor_ranges  = shift_editor_ranges(self.xr_j0, self.control_info.editor_ranges)
            self.logger.info('editor_ranges have been initialized to %s with xr_j0=%d', self.editor_ranges, self.xr_j0)

        self.specpanel_list = []

        if redraw:
            self.refresh_figs()
            self.refresh_spec_panels(restore=True, hints_dict=hints_dict)
            self.panel_frame.update()

    def get_decomp_score( self ):
        score = self.decomp_result.compute_synthetic_score(self.ignorable_flags, self.select_matrix)
        self.logger.info('score on %s is %g' % (self.model_name, score))
        return score

    def refresh_spec_panels(self, restore=False, hints_dict=None):
        if len(self.specpanel_list) > 0:
            for specpanel in self.specpanel_list:
                specpanel.destroy()
            self.specpanel_list = []

        active_peak_no = 0
        row_no_base = 0
        for k in range(self.num_eltns):
            # print('editor_ranges[%d]=' % k, self.editor_ranges[k])
            if not self.ignorable_flags[k] and len(self.editor_ranges[k]) > 0:
                active_peak_no += 1

            range_list = self.editor_ranges[k]
            panel = SpecPanel(self.ppframe,
                            editor=self,
                            j_min=self.xr_j0,
                            j_max=self.xr_j0 + len(self.x) - 1,
                            peak_no=k+1,
                            active_peak_no=None if self.ignorable_flags[k] else active_peak_no,
                            row_no_base=row_no_base,
                            col_widths=self.col_widths,
                            select_list=self.select_matrix[k],
                            range_list=range_list,
                            model=self.model,
                            # params_controllable=self.params_controllable,
                            params_controllable = False,    # controlable params are decprecated
                            opt_rec=self.opt_recs[k],
                            hints_dict=hints_dict,
                            )
            if not self.ignorable_flags[k]:
                row_no_base += len(range_list)
            if restore:
                panel.constraints_restore()
            panel.grid(row=k, column=0, sticky=Tk.N+Tk.S)
            self.specpanel_list.append(panel)

    def get_param_hints_dict(self):
        from collections import OrderedDict

        hints_dict = OrderedDict()
        for specpanel in self.specpanel_list:
            try:
                xkey = specpanel.peak.get_xkey()
                hints = specpanel.get_tau_hints()
                hints_dict[xkey] = hints
            except:
                # temporary fix: hints_dict will not be used anyway
                pass
        return hints_dict

    def log_info_for_reset(self, modification="last applied"):
        self.logger.info(DECOMP_EDITOR_LOG_HEADER)

        self.logger.info(modification + " model_name=" + self.model_name)

        self.logger.info("element cuves:")
        for k, panel in enumerate(self.specpanel_list):
            tau_hints = panel.get_tau_hints()
            xray_evaluator = self.opt_recs[k].evaluator
            param_text = xray_evaluator.get_all_params_string()
            xray_h = xray_evaluator.get_param_value(0)
            uv_evaluator = self.opt_recs_uv[k].evaluator
            uv_h = uv_evaluator.get_param_value(0)
            uv2xray_scale = uv_h/xray_h
            self.logger.info( str([k+1]) + ' ' + param_text + '; with tau constraints ' + str(tau_hints) + ' and uv/xray h ratio %g' % uv2xray_scale )

        self.logger.info("used ranges:")
        for k, panel in enumerate(self.specpanel_list):
            if self.ignorable_flags[k] or len(self.editor_ranges[k]) == 0:
                continue

            curves = []
            for m, flag in enumerate(self.select_matrix[k]):
                if flag:
                    curves.append(m+1)

            info = panel.get_info_for_reset()
            self.logger.info(str([k+1]) + ' ' + str(info[1]) + ' with element curves ' + str(curves) )

        self.logger.info(DECOMP_EDITOR_LOG_TRAILER)

    def on_mpl_button_press(self, event):
        if event.xdata is None:
            return

        if event.button == 3:
            from molass_legacy.KekLib.PopupMenuUtils import post_popup_menu
            self.create_popup_menu()
            post_popup_menu(self.popup_menu, self.mpl_canvas1_widget, event, mpl_event=True)
            return

    def create_popup_menu(self):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu( self, tearoff=0 )
            self.popup_menu.add_command( label='Save the curves', command=self.save_the_curves )

    def save_the_curves(self):
        print('save_the_curves')
        from molass_legacy.SerialAnalyzer.CurveSaverDialog import CurveSaverDialog

        curve_list = []
        for k, rec in enumerate(self.the_curves):
            label, x, y = rec
            curve = np.array([x, y]).T
            curve_list.append([label, curve, label + '.dat'])

        model_type = self.model_name.lower()
        data_type = 'xray' if self.elution_fig_type == 0 else 'uv'
        save_folder = '/'.join( [get_setting('analysis_folder'), model_type, data_type] )
        dialog = CurveSaverDialog(self.dialog, curve_list, save_folder)
        dialog.show()

    def update_range(self, pno, ad, f, t):
        print('update_range:', pno, ad, f, t, len(self.specpanel_list))
        range_ = self.editor_ranges[pno][ad]
        range_[0] = f
        range_[1] = t
        panel = self.specpanel_list[pno]
        panel.update_range(ad, f, t)
        # self.update_figs(pno)
        self.refresh_figs()

    def get_extrapolation_solver(self):
        """
        always construct a new solver to initialize simply
        """
        from molass_legacy.Extrapolation.ExtrapolationSolver import ExtrapolationSolver
        pdata, popts = self.dialog.preview_frame.get_preview_data()
        self.ex_solver = ExtrapolationSolver(pdata, popts)
        return self.ex_solver

    def compute_global_fitting(self):
        from molass_legacy.Optimizer.OptimalElution import OptimalEmg, OptimalEgh, get_h_vector, select_resc
        from molass_legacy.DataStructure.SvdDenoise import get_denoised_data

        self.logger.info('computing %s global fit', self.model_name)
        D, E, qvector, ecurve = self.dialog.serial_data.get_xr_data_separate_ly()
        h_vector, rank = get_h_vector(self.opt_recs)
        fit_recs = select_resc(h_vector, rank, self.opt_recs)
        M = get_denoised_data(D, rank=rank)

        if self.model_name[0:3] == "EMG":
            opimizer = OptimalEmg(M, ecurve, fit_recs, rank, error=E, debug=False)
        else:
            opimizer = OptimalEgh(M, ecurve, fit_recs, rank, error=E, debug=False)

        return UnifiedDecompResult(
                        xray_to_uv=None,
                        x_curve=ecurve, x=ecurve.x, y=ecurve.y,
                        opt_recs=opimizer.fit_recs,
                        max_y_xray = ecurve.max_y,
                        model_name=opimizer.model.get_name(),
                        decomposer=None,
                        uv_y=None,
                        opt_recs_uv=None,
                        max_y_uv= None,
                        global_flag=True,
                        )

    def compute_jsd(self):
        from scipy.spatial import distance

        x_curve = self.decomp_result.x_curve
        x = x_curve.x
        y = x_curve.y.copy()
        y[y < 0] = 0

        cy_list = []
        for rec in self.decomp_result.opt_recs:
            cy_list.append(rec.evaluator(x))

        ty = np.sum(cy_list, axis=0)
        return distance.jensenshannon(ty, y)

    def update_with_result(self, decomp_result, debug=False):
        self.recompute_decomposition(decomp_result=decomp_result, debug=debug)

    def reset(self):
        self.recompute_decomposition(decomp_result=self.decomp_result_init)



"""
    RunPython.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import sys
import subprocess

class RunPythonResult:
    def __init__(self, result):
        self.returncode = result.returncode
        self.stdout = result.stdout.decode()
        self.stderr = result.stderr.decode()

def run_this_python(*args):
    result = subprocess.run(
        [sys.executable, *args],
        capture_output=True,
        )
    return RunPythonResult(result)

def run_this_python_stdout_lines(*args):
    # reference  https://qiita.com/megmogmog1965/items/5f95b35539ed6b3cfa17
    proc = subprocess.Popen([sys.executable, *args],
            shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    while True:
        line = proc.stdout.readline()
        if line:
            yield line
        if not line and proc.poll() is not None:
            break



# coding: utf-8
"""
    DiffUtils.py

    Generating and applying diffs in python
    https://stackoverflow.com/questions/2307472/generating-and-applying-diffs-in-python

"""
import difflib
_no_eol = "\ No newline at end of file"

def make_patch(a,b):
  """
  Get unified string diff between two strings. Trims top two lines.
  Returns empty string if strings are identical.
  """
  diffs = difflib.unified_diff(a.splitlines(True),b.splitlines(True),n=0)
  try: _,_ = next(diffs),next(diffs)
  except StopIteration: pass
  return ''.join([d if d[-1] == '\n' else d+'\n'+_no_eol+'\n' for d in diffs])

import re
_hdr_pat = re.compile("^@@ -(\d+),?(\d+)? \+(\d+),?(\d+)? @@$")

def apply_patch(s,patch,revert=False):
  """
  Apply unified diff patch to string s to recover newer string.
  If revert is True, treat s as the newer string, recover older string.
  """
  s = s.splitlines(True)
  p = patch.splitlines(True)
  t = ''
  i = sl = 0
  (midx,sign) = (1,'+') if not revert else (3,'-')
  while i < len(p) and p[i].startswith(("---","+++")): i += 1 # skip header lines
  while i < len(p):
    m = _hdr_pat.match(p[i])
    if not m: raise Exception("Cannot process diff")
    i += 1
    l = int(m.group(midx))-1 + (m.group(midx+1) == '0')
    t += ''.join(s[sl:l])
    sl = l
    while i < len(p) and p[i][0] != '@':
      if i+1 < len(p) and p[i+1][0] == '\\': line = p[i][:-1]; i += 2
      else: line = p[i]; i += 1
      if len(line) > 0:
        if line[0] == sign or line[0] == ' ': t += line[1:]
        sl += (line[0] != sign)
  t += ''.join(s[sl:])
  return t

def file2string(path):
    with open(path) as fh:
        text = fh.read()
    return text

def string2file(text, path):
    with open(path, "w") as fh:
        fh.write(text)



"""
    Solvers.ABC.ParameterArray.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from pyabc import Parameter

MAX_NUM_PARAMS = 100

class ParameterArray(Parameter):
    def __init__(self, rv_array):
        assert len(rv_array) <= MAX_NUM_PARAMS
        self.rv_array = np.asarray(rv_array)
        self.updated = False

    def update_dict(self):
        for i, v in enumerate(self.rv_array):
            self["p%02d" % i] = v
        self.updated = True

    def __getitem__(self, key):
        if type(key) is str:
            if not self.updated:
                self.update_dict()
            return Parameter.__getitem__(self, key)
        else:
            return self.rv_array[key]

    def get(self, key):
        if not self.updated:
            self.update_dict()
        return Parameter.get(self, key)

    def copy(self) -> "ParameterArray":
        """
        Copy the parameter.
        """
        return ParameterArray(self.rv_array)

    def get_values(self):
        return self.rv_array

def parameter_to_values(df):
    try:
        return df.get_values()
    except:
        print("dataframe_to_values: type(df)=", type(df))
        values = df.values
        print("type(values), values=", type(values), values)
        return values


# coding: utf-8
"""
    MappingParams.py

    Copyright (c) 2018-2021, SAXS Team, KEK-PF
"""
from collections            import OrderedDict
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting

MAPPER_OPT_PARAMS   = [ 'uv_baseline_opt', 'uv_baseline_type', 'uv_baseline_adjust',
                        'uv_baseline_with_bpa',
                        'xray_baseline_opt', 'xray_baseline_type', 'xray_baseline_adjust',
                        'xray_baseline_with_bpa', 'dev_allow_ratio' ]

MAPPER_SIMPLEST_VALUES = [  0, 0, 0,
                            0,
                            0, 0, 0,
                            0, 0.5 ]

XRAY_LPM_DEGREE_TEXTS   = [ None, 'linear', 'quatratic', 'spline', None, 'integral' ]
# UV_CORRECT_TYPE_TEXTS   = [ 'standard', 'shift' ]
UV_CORRECT_TYPE_TEXTS   = [ None, 'linear', '', '', 'shifted', 'integral']
UV_METHOD_NAMES = ['LPM', 'LB', '', '', '', 'integral']

def get_mapper_opt_params():
    params = []
    for param in MAPPER_OPT_PARAMS:
        params += [ (param, get_setting( param ) ) ]
    return MappingParams(params)

def set_mapper_opt_params(opt_params):
    for param in MAPPER_OPT_PARAMS:
        set_setting(param, opt_params[param])

def get_mapper_simplest_params():
    params = []
    for k, param in enumerate(MAPPER_OPT_PARAMS):
        params += [ (param, MAPPER_SIMPLEST_VALUES[k] ) ]
    return MappingParams(params)

"""
    TODO: duplicated info in mapped_info and MAPPER_OPT_PARAMS in SerialSettings
"""

class MappingParams(OrderedDict):
    def __init__(self, *args):
        OrderedDict.__init__(self, *args)

    def get_xray_correction_str(self):
        xray_baseline_opt = self.get('xray_baseline_opt')
        if xray_baseline_opt == 0:
            ret = 'no correction'
        else:
            xray_adjust_with_bpa = self.get('xray_adjust_with_bpa')
            xray_baseline_type = self.get('xray_baseline_type')
            if xray_baseline_type == 6:
                ret = 'BPA'
            else:
                ret = 'LPM(%s)' % XRAY_LPM_DEGREE_TEXTS[xray_baseline_type]
                if xray_adjust_with_bpa:
                    ret += '+BPA'
            xray_baseline_adjust = self.get('xray_baseline_adjust')
            if xray_baseline_adjust == 1:
                dev_allow_ratio = self.get('dev_allow_ratio')
                ret  += ' with adjustment(%.1g)' % dev_allow_ratio
        return ret

    def get_uv_correction_str(self):
        use_xray_conc  = get_setting('use_xray_conc')
        if use_xray_conc:
            return ''

        uv_baseline_opt = self.get('uv_baseline_opt')
        if uv_baseline_opt == 0:
            ret = 'no correction'
        else:
            uv_baseline_type = self.get('uv_baseline_type')
            correct_type = UV_CORRECT_TYPE_TEXTS[uv_baseline_type]
            ret = correct_type
            uv_baseline_adjust = self.get('uv_baseline_adjust')
            if uv_baseline_adjust == 1:
                dev_allow_ratio = self.get('dev_allow_ratio')
                ret  += ' with adjustment(%.1g)' % (1 - dev_allow_ratio)
        return ret

class MappedInfo:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

    def needs_sd_corrcetion( self ):
        return self.opt_params['xray_baseline_opt'] == 1



# coding: utf-8
"""
    SecTools.AverageMakerDialog.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
import re
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, is_empty_val
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable, BlinkingFrame
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
try:
    import molass_legacy.KekLib.CustomMessageBox         as MessageBox
except:
    import OurMessageBox            as MessageBox
from SerialDataUtils import get_xray_files
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry, num_files_in
from .DataAverager import average_impl

class AverageMakerDialog(Dialog):
    def __init__(self, parent):
        self.logger = logging.getLogger(__name__)
        self.busy = False
        self.in_folder_prev = None
        self.out_pattern_manually_changed = False
        Dialog.__init__(self, parent, "Average Maker", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        folder_entry_width = 60
        num_files_label_width = 6

        grid_row = -1

        grid_row += 1
        label = Tk.Label(iframe, text="Folder Path")
        label.grid(row=grid_row, column=1)
        label = Tk.Label(iframe, text="File Name Pattern")
        label.grid(row=grid_row, column=2, padx=10)
        label = Tk.Label(iframe, text="Files")
        label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Sample Data")
        label.grid(row=grid_row, column=0, sticky=Tk.E, padx=10)

        self.in_folder1 = Tk.StringVar()
        self.in_files1 = []
        self.name_pattern1 = Tk.StringVar()
        self.num_files1 = Tk.IntVar()
        self.in_folder1_cb = lambda: self.on_entry_in_folder(self.in_folder1, self.in_files1, self.name_pattern1, self.num_files1)
        self.in_folder1_entry = FolderEntry(iframe, textvariable=self.in_folder1, width=folder_entry_width,
                                on_entry_cb=self.in_folder1_cb)
        self.in_folder1_entry.grid(row=grid_row, column=1)
        self.in_folder1_entry.bind('<Leave>', lambda *args:self.in_folder1_cb())

        label = Tk.Label(iframe, textvariable=self.name_pattern1, justify=Tk.LEFT)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files1, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        num_frame = Tk.Frame(iframe)
        num_frame.grid(row=grid_row, column=0, columnspan=3, sticky=Tk.W)

        label = Tk.Label(num_frame, text="Number of Files to Average: ")
        label.grid(row=0, column=0, sticky=Tk.W, padx=10)
        self.num_files_to_average = Tk.IntVar()
        self.num_files_to_average.set(11)
        sbox  = Tk.Spinbox(num_frame, textvariable=self.num_files_to_average,
                          from_=1, to=19, increment=1,
                          justify=Tk.CENTER, width=6)
        sbox.grid(row=0, column=1, padx=20, pady=20)

        grid_row += 1
        label = Tk.Label(iframe, text="Averaged Data")
        label.grid(row=grid_row, column=0, sticky=Tk.E, padx=10)
        self.out_folder = Tk.StringVar()
        out_folder_entry = FolderEntry(iframe, textvariable=self.out_folder, width=folder_entry_width,
                                on_entry_cb=self.on_entry_out_folder)
        out_folder_entry.grid(row=grid_row, column=1)
        out_folder_entry.bind('<Leave>', lambda *args: self.on_entry_out_folder(on_leave=True))

        self.name_pattern3 = Tk.StringVar()
        entry = Tk.Entry(iframe, textvariable=self.name_pattern3, width=30)
        entry.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)
        self.name_pattern3_tracing = True
        self.name_pattern3.trace("w", self.name_pattern3_tracer)

        self.num_files3 = Tk.IntVar()
        num_files_label = Tk.Label(iframe, textvariable=self.num_files3, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3, sticky=Tk.W)

        grid_row += 1
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Progress")
        label.grid(row=grid_row, column=0)
        self.mpb = ttk.Progressbar(iframe,orient ="horizontal", length=600, mode="determinate")
        self.mpb.grid(row=grid_row, column=1, columnspan=3, padx=5)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        self.average_btn = Tk.Button(box, text="Make Average", width=15, command=self.make_average, state=Tk.DISABLED)
        self.average_btn.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        self._show()

    def on_entry_in_folder(self, in_folder, in_files, name_pattern, num_files):
        in_folder_ = in_folder.get()
        if is_empty_val(in_folder_) or in_folder_ == self.in_folder_prev:
            return

        self.in_folder_prev = in_folder_
        files = get_xray_files(in_folder_)
        in_files.clear()
        in_files += files
        if len(files) > 0:
            dir_, file = os.path.split(files[0])
            name_pattern.set(file)
        num_files.set( len(files) )
        if self.num_files1.get() > 0:
            self.average_btn.config(state=Tk.NORMAL)

        self.in_folder1_entry.config(fg='black')

        self.out_folder.set(in_folder_ + '/averaged')
        if len(files) > 0:
            self.on_entry_out_folder()

    def on_entry_out_folder(self, on_leave=False):
        _, in_folder1 = os.path.split(self.in_folder1.get())
        name_pattern1 = self.name_pattern1.get()
        _, out_folder = os.path.split(self.out_folder.get())

        if len(name_pattern1) > 0 and (not on_leave or self.need_out_pattern_update()):
            ret_pattern = self.generate_pattern(in_folder1, name_pattern1, out_folder)
            self.name_pattern3_tracing = False
            self.name_pattern3.set(ret_pattern)
            self.out_pattern_manually_changed = False

    def need_out_pattern_update(self):
        name_pattern3 = self.name_pattern3.get()
        # print("need_out_pattern_update: name_pattern3=", name_pattern3, "changed=", self.out_pattern_manually_changed)
        if is_empty_val(name_pattern3):
            return True
        return not self.out_pattern_manually_changed

    def name_pattern3_tracer(self, *args):
        if self.name_pattern3_tracing:
            self.out_pattern_manually_changed = True
        else:
            self.name_pattern3_tracing = True

    def make_average(self):
        out_folder = self.out_folder.get()
        if os.path.exists(out_folder):
            if num_files_in(out_folder) > 0:
                reply = MessageBox.askokcancel(
                    "Clear Folder Confirmation", 
                    "The output folder is not empty.\n"
                    "This operation will clear the output folder.\n"
                    "Are you sure to proceed?",
                    parent=self
                    )
                if not reply:
                    return

        name_pattern = self.name_pattern3.get().replace('_00000', '_%05d')
        # print('name_pattern=', name_pattern)

        self.busy = True
        clear_dirs_with_retry([out_folder])

        num_average = self.num_files_to_average.get()
        average_impl(num_average, self.in_files1, out_folder, name_pattern, dialog=self)
        self.busy = False

    def is_busy(self):
        return self.busy

    def generate_pattern(self, in_folder1, name_pattern1, out_folder):
        name_pattern_re = re.compile(r'(_\d+.*)(\.dat)$')
        sub_filename = re.sub(name_pattern_re, lambda m: m.group(1) + '_avg' + m.group(2), name_pattern1)

        return sub_filename



"""
    TrimmingResult.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy._MOLASS.Version import get_version_string
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from .TrimmingDebugUtils import trimming_result_plot_impl

class TrimmingResultDialog(Dialog):
    def __init__(self, parent, pre_recog):
        self.pre_recog = pre_recog
        self.popup_menu = None
        Dialog.__init__(self, parent, "Trimming at a glance - " + get_version_string(), visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack(padx=20, pady=10)

        fig = plt.figure(figsize=(20,9))
        gs = GridSpec(2,10)

        for i, name in enumerate(["UV", "Xray"]):
            ax = fig.add_subplot(gs[i,0])
            ax.set_axis_off()
            ax.text(0.8, 0.5, name, va="center", ha="center", fontsize=20)

        axes = []
        for i in range(2):
            axis_row = []
            for j in range(3):
                start = 1+3*j
                ax = fig.add_subplot(gs[i,start:start+3])
                axis_row.append(ax)
            axes.append(axis_row)
        axes = np.array(axes)
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.fig = fig
        self.axes = axes

        result_info = self.pre_recog.ar.result_info
        self.twinxes = trimming_result_plot_impl(fig, axes, *result_info)

        fig.tight_layout()
        self.mpl_canvas.draw()
        if get_dev_setting("enable_dnd_debug"):
            self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def buttonbox(self):
        frame = Tk.Frame(self)
        frame.pack(fill=Tk.X)
        tframe = Tk.Frame(frame)
        tframe.pack(side=Tk.LEFT, padx=20)
        bframe = Tk.Frame(frame)
        bframe.pack(side=Tk.RIGHT, padx=20)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        w = Tk.Button(bframe, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def on_figure_click(self, event):
        if event.button == 1:
            return
        elif event.button == 3:
            if event.inaxes is not None:
                self.show_popup_menu(event)
                return

    def show_popup_menu(self, event):
        from molass_legacy.KekLib.TkUtils import split_geometry
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0 )
            self.popup_menu.add_command(label="Debug DnD", command=self.show_debug_dnd)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu.post(cx + int(event.x), cy + h - int(event.y))

    def show_debug_dnd(self):
        from .TrimmingDebugDnd import TrimmingDebugDnd
        print("show_debug_dnd")
        dnd = TrimmingDebugDnd(self)

    def get_trimming_info(self):
        return self.pre_recog.ar.result_info[4]

    def redraw(self, alt_info):
        print("redraw: alt_info=", alt_info)

        for axes_row in self.axes:
            for ax in axes_row:
                ax.cla()
        for ax in self.twinxes:
            ax.remove()

        result_info = self.pre_recog.ar.result_info
        self.twinxes = trimming_result_plot_impl(self.fig, self.axes, *result_info, alt_info=alt_info)
        self.mpl_canvas.draw()



"""
    SigmoidDemo.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import curve_fit
import molass_legacy.KekLib.DebugPlot as plt
from .Sigmoid import *

def basic_demo():
    plt.push()
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
    ax1.set_title("Basic Sigmoids (1, 100, k, 0)", fontsize=16)
    ax2.set_title("Extended Sigmoids (-1, 100, k, 0, -0.001, 0.001)", fontsize=16)

    x = np.arange(200)
    for k in [0.1, 0.2, 0.5, 1]:
        p1 = 1, 100, k, 0
        ax1.plot(x, sigmoid(x, *p1), label="k=%g" % k)
    ax1.legend()

    for k in [0.1, 0.2, 0.5, 1]:
        p2 = -1, 100, k, 0, -0.001, 0.001
        ax2.plot(x, ex_sigmoid(x, *p2), label="k=%g" % k)

    ax2.legend()
    fig.tight_layout()
    plt.show()
    plt.pop()

def real_demo(x_curve, a_curve, a_curve2, fit_chain=False):
    # from molass_legacy.KekLib.SciPyCookbook import smooth
    from molass_legacy.Elution.CurveUtils import simple_plot
    from .PeakRegion import PeakRegion
    from .FlowChangeCandidates import get_largest_gradients

    peak_region = PeakRegion(x_curve, a_curve, a_curve2)

    x = a_curve2.x
    # y = smooth(a_curve2.y)
    y = a_curve2.y

    gy, pp3 = get_largest_gradients(y, 3, peak_region)
    print("pp3=", pp3)
    x0 = x[pp3[0]]

    stop1 = a_curve.peak_info[0][0]
    if abs(stop1 - x0) < 5:
        # as in 20170301
        stop1 = x0 + 5
    sliceL = slice(0, stop1)
    sliceR = slice(a_curve.peak_info[-1][-1], None)

    results = []
    for slice_ in [sliceL, sliceR]:

        x_ = x[slice_]
        y_ = y[slice_]

        try:
            if x0 < x_[0] or x0 > x_[-1]:
                _, pp3_ = get_largest_gradients(y_, 3, peak_region)
                x0 = x_[pp3_[0]]
            params0 = guess_bent_sigmoid(x_, y_, x0)

            if False:
                plt.push()
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
                simple_plot(ax1, a_curve)
                ax2.plot(x, y, color="gray", alpha=0.3)
                ax2.plot(x_, y_)
                ax2.plot(x_, bent_sigmoid(x_, *params0))
                fig.tight_layout()
                plt.show()
                plt.pop()

            popt, pcov = curve_fit(bent_sigmoid, x_, y_, params0)
            L = popt[0]
            error_ratio = abs(np.std(bent_sigmoid(x_, *popt) - y_)/L)
            print("error_ratio=", error_ratio)
            if error_ratio < 0.3:
                results.append((x_, y_, params0, popt))
        except Exception as exc:
            print(exc)

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18, 5))
        simple_plot(ax1, a_curve)
        ax2.plot(x, y)
        for x_, y_, params0, popt in results:
            ax2.plot(x_, ex_sigmoid(x_, *params0), label="initial guess")
            ax2.plot(x_, ex_sigmoid(x_, *popt), label="fit result")
        ax2.legend()
        ax3.plot(x, gy)
        for k, p in enumerate(pp3):
            ax3.plot(p, gy[p], "o", label=str(k))
        ax3.legend()

        fig.tight_layout()
        plt.show()

    if fit_chain:
        from .SigmoidChain import SigmoidChain
        from DataUtils import get_in_folder

        chain = SigmoidChain(x, y, pp3)
        in_folder = get_in_folder()
        with plt.Dp():
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18, 5))
            fig.suptitle("Sigmoid Chain Demo for %s" % in_folder, fontsize=20)
            ax1.set_title("Elution Curve at λ=280", fontsize=16)
            ax2.set_title("Elution Curve at λ=340", fontsize=16)
            ax3.set_title("Gradient Curve at λ=340", fontsize=16)
            simple_plot(ax1, a_curve)
            ax2.plot(x, y, alpha=0.5)
            # ax2.plot(x, chain(x), label="fit result")
            chain.plot_segments(ax2, label="segment-%d", lw=3)
            ax2.legend()
            ax3.plot(x, gy)
            for k, p in enumerate(pp3):
                ax3.plot(p, gy[p], "o", label=str(k))
            ax3.legend()

            fig.tight_layout()
            plt.show()



"""
    Theory.SfBound.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from .JsPedersen1997 import F1
from .DjKinning1984 import P0, S0
import molass_legacy.KekLib.DebugPlot as plt

def S0_bound(q, R, K=1):
    return  1 - K * 3 / (q*R)**2

def demo():
    from .SolidSphere import get_boundary_params_simple

    qv = np.linspace(0.005, 0.3, 100)
    Rg = 35
    R = np.sqrt(5/3)*Rg

    p = P0(qv, R)
    s_ = np.ones(len(qv))
    s0 = S0(qv, R)

    b1, b2, k = get_boundary_params_simple(Rg)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,7))
    fig.suptitle("Structure Factor Bounds", fontsize=20)

    ax1.set_title("Intensities: I(q)=P(q)*S(q)", fontsize=16)
    ax2.set_title("Structure Factors: S(q)", fontsize=16)

    ax1.set_yscale('log')

    label = 'S=1'
    ax1.plot(qv, p, label=label)
    ax2.plot(qv, s_, label=label)
    label = 'Debye (K=1)'
    ax1.plot(qv, p*s0, label=label)
    ax2.plot(qv, s0, label=label)
    label = 'Fournet (K=1)'

    for i, ax in enumerate([ax1, ax2]):
        ymin, ymax = ax.get_ylim()
        if i == 1:
            ymax = 1.8
        ax.set_ylim(ymin, ymax)
        for j, b in enumerate([b1, b2]):
            ax.plot([b, b], [ymin, ymax], ':', color='gray', label='b%d' % (j+1))

    ax2.plot(qv, S0_bound(qv, 2*R), ':', color='red', label='SF Lower Bound')
    ax2.plot(qv, S0_bound(qv, 2*R, K=-1), ':', color='red', label='SF Upper Bound')

    ax1.legend()
    ax2.legend()
    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()



"""
    PropOptMenu.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from MenuButton import MenuButton

class PropOptMenu(Tk.Frame):
    def __init__(self, parent, dialog):
        Tk.Frame.__init__(self, parent)
        self.menu = MenuButton(self, "VP Analysis", [
                            ("using EGH", lambda: dialog.show_vp_analysis("EGH")),
                            ("using EDM", lambda: dialog.show_vp_analysis("EDM")),
                            ("using STC", lambda: dialog.show_vp_analysis("STC")),
                            ])
        self.menu.entryconfig(2, state=Tk.DISABLED)
        self.menu.pack()

    def config(self, **kwargs):
        self.menu.config(**kwargs)



# coding: utf-8
"""
    MXD.FastDecomposer.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.RobustPeaks import RobustPeaks
from molass_legacy.Peaks.ElutionModels import egh, compute_egh_params
from Prob.GaussianMixture import hist_to_source
from Prob.EghMixture import EghMixture
from Prob.EghMixtureUtils import get_components

def compute_egh_params_with_h(y, init_params, moments):
    h = np.max(y)
    tR, sigma, tau = compute_egh_params(init_params[1:], moments)
    return h, tR, sigma, tau

def compute_moments(x, y):
    W = np.sum(y)
    M1 = np.sum(x*y)/W              # row moment
    M2 = np.sum(y*(x-M1)**2)/W      # central moment
    M3 = np.sum(y*(x-M1)**3)/W      # central moment
    return M1, M2, M3

def verify_moments_estimation():
    """
        slightly not accurate for tau < 0
    """
    for params in [
            (1, 200, 50, 10),
            (1, 200, 50, -10)
            ]:

        h, tR, sigma, tau = params
        x = np.arange(500)
        y = egh(x, *params)

        M = compute_moments(x, y)
        init_params = h+0.1, tR-1, sigma+1, tau*1.1
        new_params = compute_egh_params_with_h(y, init_params, M)
        y_ = egh(x, *new_params)

        for p in [params, init_params, new_params]:
            print(p)

        plt.push()
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x, y_)
        fig.tight_layout()
        plt.show()
        plt.pop()

def get_safe_slice(x, params):
    tR, sigma, tau = params[1:]
    n = 3
    start = bisect_right(x, tR - n*sigma + min(0, tau))
    stop = bisect_right(x, tR + n*sigma + max(0, tau))
    return slice(start, stop)

VERY_SMALL_VALUE = 1e-5

def str_params_list(params_list):
    ret = '[\n'
    for params in params_list:
        ret += '  [' + ','.join(["%.4g" % p for p in params] ) + '],\n'
    ret += ']\n'
    return ret

class FastDecomposer:
    def __init__(self, num_components=None):
        self.num_components = num_components

    def get_initial_params(self, x, y):
        rp = RobustPeaks(x, y, debug=False)
        params_list = []
        h_list = []
        peaks = rp.get_peaks()
        for k, peak in enumerate(peaks):
            print([k], peak)
            ls, pt, rs = peak
            h = y[pt]
            h_list.append(h)
            tR = x[pt]
            sigma = (x[pt] - x[ls] + x[rs] - x[pt])/2
            tau = 0
            params_list.append((h, tR, sigma, tau))
        if self.num_components is None:
            pass
        else:
            num_diff = len(params_list) - self.num_components
            if num_diff == 0:
                pass
            elif num_diff < 0:
                new_params_list = []
                assert len(peaks) == 1
                for peak, h, params in zip(peaks, h_list, params_list):
                    ls, pt, rs = peak
                    lsw = x[pt] - x[ls]
                    rsw = x[rs] - x[pt]
                    if lsw > rsw:
                        side_list = [-1, 0, 1]
                    else:
                        side_list = [0, 1, -1]
                    for k in range(-num_diff+1):
                        side = side_list[k]
                        if side == 0:
                            new_params_list.append(params)
                        elif side < 0:
                            h_ = y[ls]
                            tR_ = x[ls] - lsw/4
                            sigma_ = lsw/2
                            tau_ = 0
                            new_params_list.append((h_, tR_, sigma_, tau_))
                        else:
                            h_ = y[rs]
                            tR_ = x[rs] + rsw/4
                            sigma_ = rsw/2
                            tau_ = 0
                            new_params_list.append((h_, tR_, sigma_, tau_))

                params_list = sorted(new_params_list, key=lambda x: x[1])
                h_list = [p[0] for p in params_list]

        return rp.max_y, h_list, params_list

    def fit(self, x, y):
        max_y, h_list, params_list = self.get_initial_params(x, y)

        primary = np.argmax(h_list)
        print("init_params_list=", str_params_list(params_list))

        my2 = max_y**2
        last_r2 = None
        last_params_list = None
        for i in range(100):
            sigma_boundary = params_list[primary][2]
            cy_list = []
            ty = np.zeros(len(x))
            for params in params_list:
                y_ = egh(x, *params)
                cy_list.append(y_)
                ty += y_

            r2 = np.average((ty - y)**2)/my2
            if last_r2 is not None:
                if r2 > last_r2:
                    params_list = last_params_list
                    break

            last_r2 = r2
            # temp fix
            ty[np.abs(ty)==0] = VERY_SMALL_VALUE

            weights = []
            new_params_list = []
            for h_init, cy, params in zip(h_list, cy_list, params_list):
                h = params[0]
                w = cy/ty
                weights.append(w)
                wy = w*y
                if i < 50:
                    slice_ = get_safe_slice(x, params)
                    M_ = compute_moments(x[slice_], wy[slice_])
                else:
                    M_ = compute_moments(x, wy)
                params = list(compute_egh_params_with_h(wy, params, M_))
                params[2] = min(sigma_boundary, params[2])
                new_params_list.append(params)

            last_params_list = params_list
            params_list = new_params_list

            if False and i%10 == 0:
                print([i], 'r2=', r2)
                print([i], 'params_list=', str_params_list(params_list))
                print([i], 'new_params_list=', str_params_list(new_params_list))

                plt.push()
                fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))
                ax1.set_title("step %d" % i)
                ax1.plot(x, y)
                for y_ in cy_list:
                    ax1.plot(x, y_, ':')
                ax1.plot(x, ty, ':', color='red')
                ax2.plot(x, y)

                for w in weights:
                    y_ = w*y
                    ax2.plot(x, y_, ':')

                ax3.plot(x, y)
                ty_ = np.zeros(len(x))
                for params in new_params_list:
                    y_ = egh(x, *params)
                    ax3.plot(x, y_, ':')
                    ty_ += y_

                ax3.plot(x, ty_, ':', color='red')

                fig.tight_layout()
                plt.show()
                plt.pop()


        self.x = x
        self.y = y
        self.params_list = params_list

    def get_components(self):
        x = self.x
        ty = np.zeros(len(x))
        cy_list = []
        for params in self.params_list:
            cy = egh(x, *params)
            cy_list.append(cy)
            ty += cy
        return cy_list, ty



# coding: utf-8
"""
    ExtrapolationDebugger.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from DataDebugger import DataDebugger

class ExtrapolationDebugger(DataDebugger):
    def __init__(self):
        DataDebugger.__init__(self, "ex-debugger-00")

    def save_info(self, M, C, A, param_info):
        sub_folder = self.get_sub_folder()
        np.savetxt(sub_folder + '/M.dat', M)
        np.savetxt(sub_folder + '/C.dat', C)
        np.savetxt(sub_folder + '/A.dat', A)
        with open(sub_folder + '/param_info.txt', 'w') as fh:
            fh.write(str(param_info) + '\n')



"""
    SecTheory.MomentAnalysis.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import root
from molass_legacy.Peaks.ElutionModels import compute_moments
from .BasicModels import robust_single_pore_pdf_scaled as elutionmodel_func
import molass_legacy.KekLib.DebugPlot as plt

def estimate_sec_params(ecurve, debug=False):
    x = ecurve.x
    y = ecurve.y

    M = compute_moments(x, y)
    m = M[0]
    s = np.sqrt(M[1])

    """
        t0, n, t
    """

    def fun(x):
        return [
            x[0] + x[1]*x[2] - M[0],
            2*x[1]*x[2]**2 - M[1],
            6*x[1]*x[2]**3 - M[2],
            ]

    t0_init = 0
    n_init = M[0]
    t_init = 1

    ret = root(fun, (t0_init, n_init, t_init))

    if debug:
        print("ret.success=", ret.success)
        print("ret.message=", ret.message)
        print("ret.x=", ret.x)
        t0, np_, tp_ = ret.x

        with plt.Dp():
            fig, ax = plt.subplots()

            ax.plot(x, y)

            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for px in [m-s, m, m+s]:
                ax.plot([px, px], [ymin, ymax], color="green")

            ax.plot(x, elutionmodel_func(x - t0, np_, tp_))

            fig.tight_layout()
            plt.show()



"""
    Optimizer.RgCurveComputer.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
from bisect import bisect_right
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.Optimizer.TheUtils import get_optimizer_folder

class RgCurveComputerDialog(Dialog):
    def __init__(self, parent, dialog, jobno):
        self.logger = logging.getLogger(__name__)
        self.logger.info("RgCurveComputerDialog opened for job %s.", jobno)
        self.parent = parent
        self.dialog = dialog
        self.fullopt = dialog.fullopt
        self.params = dialog.canvas.get_best_params()
        self.sd = dialog.optinit_info.sd
        self.jobno = jobno
        self.applied = False
        Dialog.__init__(self, parent, "RgCurveComputer", visible=False)

    def show(self):
        # self.after(3000, self.close)
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        fig, axes = plt.subplots(ncols=2, figsize=(12,5))
        self.fig = fig

        self.axes = axes
        self.draw_recomputed_rg_curve()
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

    def draw_recomputed_rg_curve(self):
        fullopt = self.fullopt
        fv, score_list, Pxr, Cxr, Puv, Cuv, mapped_UvD = fullopt.objective_func(self.params, return_full=True)

        rg_curve = fullopt.rg_curve
        D, E, qv, xr_curve = self.sd.get_xr_data_separate_ly()

        ax1, ax2 = self.axes
        self.axt = axt = ax1.twinx()
        axt.grid(False)

        nc = fullopt.n_components

        fig = self.fig
        fig.suptitle("Recomputed Rg Curve Proof Plot", fontsize=20)

        ax1.set_title("XR Elution Components", fontsize=16)
        x = xr_curve.x
        y = xr_curve.y
        ax1.plot(x, y, color="orange")
        for k, cy in enumerate(Cxr):
            color = "red" if k == nc - 1 else None
            ax1.plot(x, cy, ":", color=color)

        # Guinier Plot
        i = bisect_right(qv, 0.05)
        gslice = slice(0, i)
        ax2.set_title("XR Scatteing Components Guinier Plot", fontsize=16)
        qv2 = qv[gslice]**2
        for k, y in enumerate(Pxr.T):
            color = "red" if k == nc - 1 else None
            ax2.plot(qv2, np.log(y[gslice]), color=color)

        fig.tight_layout()
        fig.subplots_adjust(top=0.82)

    def apply(self):
        self.applied = True
        self.dummy_simulation()

    def dummy_simulation(self):
        import shutil
        from molass_legacy.Optimizer.OptDataSets import get_current_rg_folder

        optimizer_folder = get_optimizer_folder()
        job000_folder = os.path.join(os.path.join(optimizer_folder, "jobs"), "000")
        src_rg_folder = get_current_rg_folder(possibly_relocated=True, current_folder=job000_folder)
        current_folder = os.path.join(os.path.join(optimizer_folder, "jobs"), self.jobno)
        tgt_rg_folder = get_current_rg_folder(compute_rg=True, possibly_relocated=True, current_folder=current_folder)
        shutil.copytree(src_rg_folder, tgt_rg_folder)
        self.logger.info("copied for testing from %s to %s", src_rg_folder, tgt_rg_folder)

    def close(self):
        self.logger.info("RgCurveComputerDialog closed.")
        self.cancel()



"""
    Models/Stochastic/PsdEstimate.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import seaborn as sns
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.EGH import egh
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.Models.ModelUtils import plot_component_curves, get_paired_ranges_from_params_array
from molass_legacy.KekLib.ContextUtils import get_variables_from_context
from molass_legacy.Models.EGH import EGH, EGHA

sns.set_theme()

def psd_estimate_by_paired_ranges(folder, batch, context, affine=True, with_cf=False, separate_ratio=None,  debug=False, count=0,
                                  computed_params=None, compute_residual_ratio=False, use_lrf_src=False):
    logger = logging.getLogger(__name__)
    logger.info("estimating with psd_estimate_by_paired_ranges")

    uv_x, uv_y, uv_peaks = get_variables_from_context(context, 'uv_x, uv_y, uv_peaks')
    baselines, = get_variables_from_context(context, 'baselines')   # note the results are returned as a list
    xr_x, xr_y, xr_peaks = get_variables_from_context(context, 'xr_x, xr_y, xr_peaks')

    if False:
        with plt.Dp(button_spec=["OK", "Cancel"]):
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10, 4))
            fig.suptitle("Impl Plot for %s" % get_in_folder(folder))
            plot_component_curves(ax1, uv_x, uv_y, uv_peaks, baselines[0], color='blue', affine=affine)
            plot_component_curves(ax2, xr_x, xr_y, xr_peaks, baselines[1], color='orange', affine=affine)
            fig.tight_layout()
            ret = plt.show()
        if not ret:
            return

    sd, = get_variables_from_context(context, 'sd')

    from Selective.LrfSource import LrfSource
    lrf_src_args1 = [uv_x, uv_y, xr_x, xr_y, baselines]
    peak_params_set = batch.get_peak_params_set()
    lrf_src = LrfSource(sd, batch.corrected_sd, lrf_src_args1, peak_params_set)
    return pds_estimate_impl(folder, lrf_src, compute_residual_ratio=compute_residual_ratio)

def pds_estimate_impl(folder, lrf_src, count=0, compute_residual_ratio=False, debug=False):
    from importlib import reload
    import Models.Stochastic.MomentsStudy
    reload(Models.Stochastic.MomentsStudy)
    from molass_legacy.Models.Stochastic.MomentsStudy import moments_study_impl
    from molass_legacy.Models.Stochastic.LognormalPoreColumn import compute_residual_ratio_impl
    from molass_legacy.Models.Stochastic.LnporeUtils import plot_lognormal_fitting_state

    lnp_params, rgs = moments_study_impl(lrf_src, return_rgs=True, debug=debug)
    x = lrf_src.xr_x
    y = lrf_src.xr_y
    residual = compute_residual_ratio_impl(x, y, rgs, lnp_params, debug=False)
    if compute_residual_ratio:
        return residual

    print("[lnp_params, [residual]]=", [lnp_params, [residual]])
    pore_params = np.concatenate([lnp_params, [residual]])
    title = "PSD Estimate Result for " + get_in_folder(folder)
    plot_lognormal_fitting_state(x, y, lnp_params, rgs, plot_boundary=True, title=title, save_fig_as='temp/fig-%03d.png' % count)
    return pore_params


"""
    AbnormalityCheck.py

    Copyright (c) 2017-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, get_xray_picking
import molass_legacy.KekLib.CustomMessageBox as MessageBox

PLOT_GRADIENT   = True
DEGUB_PLOT      = False

BUBLE_MAX_LENGTH        = 5
ABNORMAL_GRAD_BOUNDARY  = 0.2
TAIL_CHECK_WIDTH        = 10
TAIL_NEG_RATIO          = 0.5

# ------------------------------------------------------------------------------
#   Detect abnormal points caused by bubbles
# ------------------------------------------------------------------------------
def bubble_check(serial_data, **kwargs):
    from molass.DataUtils.AnomalyHandlers import bubble_check_impl
    serial_data.wait_until_ready()
    exclude = []

    if not serial_data.is_serial():
        return exclude

    y   = serial_data.xray_curve.y_orig     # use original considering such a case as 20171226
    # gy  = serial_data.xray_curve.gy
    return bubble_check_impl(y, **kwargs)

# ------------------------------------------------------------------------------
#   Exclusion Management
# ------------------------------------------------------------------------------
def exclude_abnormality(serial_data, file_info_table, logger, dialog=None, parent=None, quiet=True):
    if parent is None:
        parent = dialog

    excluded = file_info_table.get_excluded_indeces()
    excluded_spec = None
    if len( excluded ) > 0:
        excluded_ = [ str(i) for i in excluded ]
        reply = MessageBox.askokcancel(
                "Exclude Confirmation",
                'Rows № ' + ','.join( excluded_ )
                            + " are marked to be excluded.\n"
                            + "Those excluded rows will be interpolated instead.\n"
                            + "Ok?",
                parent=parent,
                )
        if not reply:
            return
        excluded_spec = 1
    else:
        ask = True if get_setting('data_exclusion') else False
        exclude = bubble_check(serial_data)
        if len(exclude) > 0:
            excluded_spec = 2
            file_info_table.set_exclude( exclude )
            excluded = exclude

    if len( excluded ) > 0:
        serial_data.exclude_intensities( excluded )
        if excluded_spec == 2:
            if not quiet:
                message = "Xray data at elution points " + str(excluded) + " have been removed and interpolated."
                MessageBox.showinfo(
                    "Exclusion Done", message,
                    parent=parent,
                    )

def update_abnormality_fix_state(sd, file_info_table, logger, dialog=None):
    to_be_excluded = file_info_table.get_excluded_indeces()
    excluded_set = list(sd.excluded_set)
    if excluded_set == to_be_excluded:
        return

    logger.info("updating data exclusion with inconsistent sets: %s vs. %s.", str(to_be_excluded), str(sd.excluded_set))
    sd.exclude_intensities(to_be_excluded)

    if dialog is not None:
        logger.info("updating data exclusion for the original data along with the figures.")
        dialog.serial_data.exclude_intensities(to_be_excluded)
        dialog.fig_frame.draw_figure()
        dialog.update()



# coding: utf-8
"""
    OurSubprocess.py

    Copyright (c) 2016-2018, Masatsuyo Takahashi, KEK-PF
"""
import sys
from subprocess     import Popen, PIPE

def set_mode_and_get_flags():
    # learned at
    # https://www.activestate.com/blog/2007/11/supressing-windows-error-report-messagebox-subprocess-and-ctypes
    if sys.platform.startswith("win"):
        # Don't display the Windows GPF dialog if the invoked program dies.
        # See comp.os.ms-windows.programmer.win32
        #  How to suppress crash notification dialog?, Jan 14,2004 -
        #     Raymond Chen's response [1]
        import ctypes
        SEM_NOGPFAULTERRORBOX   = 0x0002    # From MSDN
        ctypes.windll.kernel32.SetErrorMode(SEM_NOGPFAULTERRORBOX);
        CREATE_NO_WINDOW  =0x08000000       # From MSDN
        subprocess_flags = CREATE_NO_WINDOW
    else:
        subprocess_flags = 0

    return subprocess_flags

def exec_subprocess( cmd, shell=False ):
    subprocess_flags = set_mode_and_get_flags()

    p = Popen( cmd, shell=shell, stdout=PIPE, stderr=PIPE,
                universal_newlines=True,
                creationflags=subprocess_flags )
    out, err = p.communicate()

    return out, err



"""
    Simulative/SimulativeDemo.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import seaborn as sns
sns.set_theme()

def demo(parent=None):
    from importlib import reload
    import Simulative.LognormalPsd
    reload(Simulative.LognormalPsd)
    from Simulative.LognormalPsd import lognormalpore_model_interactive_impl
    """
    these parameter values were taken from 20190529_1
    """
    x = np.arange(330)
    N = 1000
    T = 0.7327
    t0 = -143.8
    me = 1.5
    mp = 1.5
    mu = 6.197
    sigma = 0.05
    scales = np.array([2.312, 10, 2.07, 0.6056])
    params = np.concatenate([[N, T, t0, me, mp, mu, sigma], scales])
    rgs = np.array([91.40615662, 67.23922297, 47.23719934, 35.73305675])
    lognormalpore_model_interactive_impl(x, None, params, rgs, use_ty_as_data=True, plot_mnp=True, window_title="Demo", parent=parent)

if __name__ == "__main__":
    import os
    import sys
    this_dir = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(os.path.join(this_dir, '..'))
    demo()


# coding: utf-8
"""
    RankAnalysis.py

    custom class of opticspy

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from SvdDenoise import get_denoised_data
import molass_legacy.KekLib.DebugPlot as plt

class RankAnalysis:
    def __init__(self, data):

        scores = []
        for rank in range(1, 9):
            D = get_denoised_data(data, rank=rank)
            scores.append(np.linalg.norm(D - data))

        plt.push()
        plt.plot(scores, '-o')
        plt.show()
        plt.pop()



"""
    KekLib/ContextUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def get_variables_from_context(context, variables_str):
    """
    Get variables from context.
    fully generated by Copy AI.
    """
    variables = variables_str.split(',')
    variables = [v.strip() for v in variables]
    return [context[v] for v in variables]



"""
    Simplified version of multiprocessin.shared_memory.SharedMemory
    for debug purposes
"""

import mmap
from .OurMultiprocessing import _make_filename

ANONYMOUS_MEMORY = -1

class SharedMemory:
    def __init__(self, name=None, create=False, size=0):
        if create:
            temp_name = _make_filename() if name is None else name
            self._mmap = mmap.mmap(ANONYMOUS_MEMORY, size, tagname=temp_name)
            self._name = temp_name
        else:
            self._mmap = mmap.mmap(ANONYMOUS_MEMORY, size, tagname=name)
            self._name = name

        self._size = size
        self._buf = memoryview(self._mmap)

    def __del__(self):
        try:
            self.close()
        except OSError:
            pass

    def close(self):
        # print(__name__, 'close')
        """Closes access to the shared memory from this instance but does
        not destroy the shared memory block."""
        if self._buf is not None:
            self._buf.release()
            self._buf = None
        if self._mmap is not None:
            self._mmap.close()
            self._mmap = None

    @property
    def buf(self):
        "A memoryview of contents of the shared memory block."
        return self._buf

    @property
    def name(self):
        "Unique name that identifies the shared memory block."
        reported_name = self._name
        return reported_name

    @property
    def size(self):
        "Size in bytes."
        return self._size



"""
    StcAdjuster.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt

class StcAdjuster:
    def __init__(self):
        pass

    def fit(self, fullopt, fv_array, x_list, debug=False):
        params = x_list[0]
        print(fv_array.shape, len(x_list), params.shape)

        n_iterations = 10   # any size > 0

        n = fullopt.n_components
        nc = n - 1
        i = nc + 2 + nc + 1                 # b
        j = nc + 2 + nc + 2 + nc + 7 + 2    # t0
        fv = fv_array[0]

        fit_params = params.copy()

        def objective(p):
            fit_params[[i,j]] = p
            fullopt.prepare_for_optimization(fit_params)
            return (fullopt.objective_func(fit_params) - fv)**2

        xr_curve = fullopt.xr_curve
        b, t0 = params[[i,j]]
        ret = minimize(objective, (b, xr_curve.x[0] + t0))
        fit_params[[i,j]] = ret.x

        if debug:
            print("b : %g → %g" % (b, ret.x[0]))
            print("t0: %g → %g" % (t0, ret.x[1]))
            with plt.Dp():
                fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(12,10))
                for i, p in enumerate([params, fit_params]):
                    ax1 = axes[i,0]
                    ax2 = axes[i,1]
                    fullopt.prepare_for_optimization(p)
                    fullopt.objective_func(p, plot=True, axis_info=(fig, (ax1, ax2, None, None)))
                fig.tight_layout()
                plt.show()

        self.fit_ij = [i,j]
        self.fit_delta = ret.x - params[[i,j]]

    def convert(self, params):
        if type(params) is list:
            params = [self.convert(p) for p in params]
        else:
            params = params.copy()
            for k in range(2):
                params[self.fit_ij[k]] += self.fit_delta[k]
        return params



"""
    UvCorrectorEgh.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import linregress
from scipy.optimize import curve_fit
from molass_legacy.Peaks.EghSupples import d_egh
from .UvCorrector import UvCorrector

class UvCorrectorEgh(UvCorrector):
    def __init__(self, curve1, curve2):
        self.curve1 = curve1
        self.curve2 = curve2
        self.x = curve2.x
        self.y = curve2.y

    def model_curve(self):
        pass

    def correction_curve(self, x, *params):
        slope, intercept, H, tR, sigma, tau = params
        return slope * x + intercept + d_egh(x, H, tR, sigma, tau)

    def fit(self, peak_region, debug=False, fig_file=None):
        x = self.x
        y = self.y

        fit_slice = self.get_widest_possible_fit_slice(peak_region)

        fx = x[fit_slice]
        fy = y[fit_slice]
        slope, intercept = linregress(fx, fy)[0:2]
        curve1 = self.curve1
        emg_peaks = curve1.get_emg_peaks()
        n = curve1.primary_peak_no
        emg_peak = emg_peaks[n]
        opt_params = emg_peak.opt_params
        H = opt_params[0]
        tR = opt_params[1]
        sigma = opt_params[2]
        tau = opt_params[3]
        pos_p0 = np.array([slope, intercept, +H, tR, sigma, tau])
        neg_p0 = np.array([slope, intercept, -H, tR, sigma, tau])

        popts = []
        errors = []
        for k, p0 in enumerate([pos_p0, neg_p0]):
            if debug:
                from .UvCorrectorIllust import debug_plot
                debug_plot("Initial Parameters", self, x, y, p0, fit_slice, fig_file=fig_file)

            try:
                popt, pcov = curve_fit(self.correction_curve, fx, fy, p0)
                error = np.std(self.correction_curve(fx, *p0) - fy)
                if debug:
                    debug_plot("Optimized Parameters", self, x, y, popt, fit_slice, init_params=p0, fig_file=fig_file)
            except Exception as exc:
                print(exc)
                popt = None
                error = np.inf

            popts.append(popt)
            errors.append(error)

        print("errors=", errors)

        self.popt = popts[0] if errors[0] < errors[1] else popts[1]



"""
    IntensityData.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF
"""
import numpy        as np
from scipy          import stats
from molass_legacy.KekLib.NumpyUtils     import get_valid_index, np_loadtxt_robust
from HeadAnomalies  import HeadAnomalies

BASIC_CHECK_LENGTH          = 10
SLOPE_INDEX                 = 0
R_VALUE_INDEX               = 2
P_VALUE_INDEX               = 3
ACCEPTABLE_P_VALUE_LIMIT    = 0.3       # 20160615/1PCAF_00173_sub.dat, subtract_SSC/IN_c380_00001_sub.dat
ACCEPTABLE_P_RANGE_SIZE     = 4         # subtract_SSC/IN_c380_00001_sub.dat [5, 10]
ACCEPTABLE_P_MEDIAN         = 0.05      # 0.0424 for 20160615/1PCAF_00176_sub.dat
TOO_NARROW_P_MEDIAN         = ACCEPTABLE_P_MEDIAN + 0.01

ACCEPTABLE_R_VALUE_LIMIT    = 0.1
ACCEPTABLE_R_MEAN           = -0.7
TOO_NARROW_R_MEAN           = ACCEPTABLE_R_MEAN + 0.01
SUFFICIENT_R_MEAN           = -0.99
SUFFICIENTLY_LOW_ERROR      = 0.003
ACCEPTABLE_ERROR_SCORE      = 0.1
ACCEPTABLE_BASIC_QUALITY    = 0.8

"""
    p50=0.028 for 'IN_c380_00010_sub.dat'
    any other ?
"""

DEBUG = False
ACCEPTABLE_POSITIVE_RATIO   = 0.1

class IntensityData:
    def __init__( self, source, add_smoother=False ):

        self.comments = None

        if type( source ) == str:
            array_, comment_lines = np_loadtxt_robust( source, usecols=(0, 1, 2) )
            if len( comment_lines ) > 0:
                self.set_comments( comment_lines )

        elif type( source ) == np.ndarray:
            array_ = np.array( source )
        elif type( source ) == IntensityData:
            array_ = np.array( source.array )
        else:
            assert( False )

        self.orig_array = array_
        # print( 'orig_array.shape=', self.orig_array.shape )

        self.add_smoother_flag = add_smoother

        self.initialize()

    def set_comments( self, comment_lines ):
        self.comments = ''.join( comment_lines )

    def initialize( self ):
        array_ = self.orig_array

        I_ = array_[ :, 1 ]
        E_ = array_[ :, 2 ]
        # I_valid = get_valid_index( I_ )
        I_positive = np.logical_and( I_ > 0, E_ > 0 )
        array_ = array_[ I_positive, : ]
        self.positive_where = np.where( I_positive )
        self.positive_index = np.ones( ( I_.shape[0], ), dtype=int ) * ( -1 )
        for i, p in enumerate( self.positive_where[0] ):
            self.positive_index[ p ] = i

        self.array = array_
        self.positive_ratio = self.compute_positive_ratio()

        self.Q  = self.array[ :, 0 ]
        self.I  = self.array[ :, 1 ]
        self.E  = self.array[ :, 2 ]
        self.size = self.array.shape[0]
        # print( 'self.size=', self.size )
        self.X  = self.Q**2
        self.Y  = np.log( self.I )
        W_ = 1/self.E**2
        self.W  = np.sqrt( W_ / np.sum( W_ ) )
        if DEBUG:
            print( self.Q[0:5] )
            print( self.I[0:5] )
            print( self.E[0:5] )

        try:
            self.observe_basic_condition()
            self.compute_basic_range()
        except Exception as exc:
            if DEBUG:
                raise exc
            else:
                self.basic_quality = 0
                return

        self.smoother = None
        if self.add_smoother_flag:
            if self.positive_ratio >= ACCEPTABLE_POSITIVE_RATIO:
                try:
                    self.add_smoother()
                except:
                    raise RuntimeError( 'positive_ratio=%g' % self.positive_ratio )

    def has_acceptable_quality( self ):
        return self.basic_quality > 0 and self.positive_ratio >= ACCEPTABLE_POSITIVE_RATIO

    def copy( self, start=0 ):
        array_ = self.orig_array[ start:, : ]
        return IntensityData( array_, add_smoother=self.add_smoother_flag )

    def get_usable_range( self ):
        return self.f1, self.t1

    def get_approximate_range( self ):
        f = max( 10,  min( 20, self.f1 ) )
        return f, self.size // 4

    def get_guinier_valid_xy( self ):
        return self.X, self.Y, self.E

    def compute_positive_ratio( self ):
        I = self.orig_array[ :, 1 ]
        N = I.shape[0]

        y05, y95 = np.percentile( I, [ 5, 95 ] )
        if y05 > 0:
            if DEBUG: print( 'y05, y95=', y05, y95 )
            dy = ( y95 - y05 )/( N - 1 )
            weights = np.arange( y95, y05 - dy/10, -dy )
            if DEBUG: print( 'N=', N, 'weights.shape=', weights.shape )
        else:
            weights = np.ones( ( N, ) )

        # print( 'weights.shape=', weights.shape, 'N=', N )
        assert( weights.shape[0] == N )

        weights /= np.sum( weights )
        ratio = np.sum( weights[ I > 0 ] )
        if DEBUG: print( 'positive_ratio=', ratio )
        return ratio

    def observe_basic_condition( self ):
        i = 0
        at_end = False
        eval_array = []
        while i * BASIC_CHECK_LENGTH < self.X.shape[0] and not at_end:
            start   = i * BASIC_CHECK_LENGTH
            end     = start + BASIC_CHECK_LENGTH
            if end + BASIC_CHECK_LENGTH > self.X.shape[0]:
                end = self.X.shape[0]
                at_end = True
            result  = self.evaluate_range( slice( start, end ) )
            eval_array.append( result )
            i += 1

        array_ = np.array( eval_array )

        if DEBUG:
            from molass_legacy.KekLib.NumpyUtils import np_savetxt
            np_savetxt( "observe_basic_condition.csv", array_ )

        self.slope_array        = array_[:,SLOPE_INDEX]
        self.r_array = r_array  = array_[:,R_VALUE_INDEX]
        self.p_array = p_array  = array_[:,P_VALUE_INDEX]
        # print( 'p_array=', p_array[0:10] )

        self.p_acceptable = p_array < ACCEPTABLE_P_VALUE_LIMIT
        self.p_ok = p_ok = np.where( self.p_acceptable )[0]
        # print( 'p_ok=', p_ok )
        # TODO: case p_ok is empty

        t10_ok = False
        for k in range( min( 4, len( p_ok ) ) ):
            f10 = p_ok[k]

            # pick the continuous ( simply connected ) part at head
            head_scale = np.array( range( f10, f10 + p_ok.shape[0]-k ) )
            p_ok_head = np.where( p_ok[k:] == head_scale )[0]
            # print( 'p_ok_head=', p_ok_head )
            t10 = p_ok[ p_ok_head[-1] ]

            if t10 > f10 + 1:
                # print( 't10=', t10 )
                if t10 > 3:
                    t10_ok = True
                    break
                else:
                    # as in case of subtract_SSC/IN_c380_00010_sub.dat
                    continue

        if not t10_ok:
            k = min( 1, p_ok.shape[0]-1 )
            # k == 0 seems to be inappropriate
            # as for aggregation_20160524/AIMdim01_00171_sub.dat
            f10 = p_ok[k]
            t10 = f10

        if DEBUG: print( 'f10, t10=', [ f10, t10 ] )

        self.p50 = np.percentile( p_array[f10:t10+1], 50 ) if t10 >= f10 + ACCEPTABLE_P_RANGE_SIZE else TOO_NARROW_P_MEDIAN
        if DEBUG: print( 'self.p50=', self.p50 )
        self.p_value_ok = self.p50 <= ACCEPTABLE_P_MEDIAN
        if DEBUG: print( 'p50=', self.p50, ', basically_ok=', self.p_value_ok )

        self.basic_condition =  max( 0, ( ACCEPTABLE_P_MEDIAN - self.p50 ) / ACCEPTABLE_P_MEDIAN )

        self.f10 = f10
        self.t10 = t10

    def evaluate_range( self, slice_ ):
        X_ = self.X[ slice_ ]
        Y_ = self.Y[ slice_ ]
        slope, intercept, r_value, p_value, std_err = stats.linregress( X_, Y_ )
        return [ slope, intercept, r_value, p_value, std_err ]

    def get_acceptable_XYW( self ):
        pass

    def compute_basic_range( self ):

        f10 = self.f10
        t10 = self.t10

        if f10 > 0:
            f1 = None
            for j in range( (f10-1)*BASIC_CHECK_LENGTH, f10*BASIC_CHECK_LENGTH ):
                # print( 'j=', j )
                result  = self.evaluate_range( slice( j, j+BASIC_CHECK_LENGTH ) )
                if result[1] < ACCEPTABLE_P_VALUE_LIMIT:
                    f1 = j
                    break
            if f1 is None:
                f1 = f10*BASIC_CHECK_LENGTH
        else:
            f1 = 0

        t1 = None
        t1_max = min( self.X.shape[0], (t10+1)*BASIC_CHECK_LENGTH )
        for j in range( t10*BASIC_CHECK_LENGTH, t1_max ):
            # print( 'j=', j )
            result  = self.evaluate_range( slice( j, j+BASIC_CHECK_LENGTH ) )
            if result[P_VALUE_INDEX] < ACCEPTABLE_P_VALUE_LIMIT:
                t1 = j
            else:
                break

        if t1 is None:
            t1 = t1_max - 1

        head_slice = slice( 0, self.size//4 )
        Ih_ = self.I[head_slice]
        Xh_ = self.X[head_slice]
        Yh_ = self.Y[head_slice]
        ha = HeadAnomalies( Ih_, Xh_, Yh_ )
        ha.remove_head_anomalies()
        if DEBUG: print( 'ha.start_point=', ha.start_point )
        self.ha_end = ha.start_point

        f_min = ha.start_point
        # f1 = max( f1, f_min )
        f1  = f_min

        if DEBUG: print( 'p_value_ok=', self.p_value_ok )
        if self.p_value_ok:
            f_ = f1
            t_ = max( f1 + 100, t1 )
        else:
            f_ = 10
            t_ = f_ + 200

        self.basic_quality_reg = e = self.evaluate_range( slice( f_, t_+1 ) )
        self.basic_error = abs( e[-1]/e[0] )
        if DEBUG:
            print( 'f_, t_=', [ f_, t_ ] )
            print( 'slope=', e[0], ', std_err=', e[-1], ', basic=', self.basic_error )
        self.basic_quality = np.exp( - max( 0, self.basic_error - SUFFICIENTLY_LOW_ERROR )*10 )
        self.basically_ok = self.basic_quality >= ACCEPTABLE_BASIC_QUALITY

        if DEBUG: print( 'f_, t_=', [ f_, t_ ], ', basic_quality=', self.basic_quality )

        if self.basically_ok:
            f_ = f1
            t_ = max( f1 + 100, t1 )
        else:
            f_ = 10
            t_ = f_ + 200

        self.f1 = f_
        self.t1 = t_

    def add_smoother( self ):
        from GaussianProcessDep     import SmootherChain

        if DEBUG: print( 'add_smoother: basic_quality=', self.basic_quality )
        if self.basically_ok:
            # f1, t1 = self.f1, self.t1
            f1, t1 = self.get_usable_range()
            f1 = min( 30, f1 )
            if t1 - f1 < 40:
                # f1, t1 = self.get_approximate_range()
                t1 = f1 + 40
        else:
            f1, t1 = self.get_approximate_range()

        if DEBUG: print( 'add_smoother: f1, t1=', f1, t1 )

        ft_size = t1 - f1
        quarter = ft_size // 4
        half    = quarter * 2

        boundaries  = [ f1, f1 + quarter, f1 + half, t1, self.size ]
        # print( 'boundaries=', boundaries )
        if self.basically_ok:
            nuggets = [ 1e-3, 1e-3, 1e-3, 1e-3 ]
        elif self.basic_quality >= 0.3:
            nuggets = [ 1e0, 1e0, 1e0, 1e0 ]
        else:
            nuggets = [ 1e1, 1e1, 1e1, 1e1 ]

        self.smoother = SmootherChain( self.Q, self.I, boundaries=boundaries, nuggets=nuggets )
        # self.smoother = Smoother( self.Q, self.I )

        # smooth data for aggregation check
        if self.basically_ok:
            self.slice_ = slice_ = slice( f1, t1+1 )
            self.SQ = self.Q[ slice_ ]
            self.SX = self.X[ slice_ ]
            W_ = self.W[ slice_ ]**2
        else:
            self.SQ = self.Q
            self.SX = self.X
            W_ = self.W**2
        self.SI = self.smoother( self.SQ )
        self.SY = np.log( self.SI )
        self.SW = np.sqrt( W_ / np.sum( W_ ) )



"""
    NumericalUtils.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""

import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

OUTLIER_SCALE = 2
OUTLIER_IGNORE_LIMIT = 1e-6

def safe_ratios_debug_plot(x, y, xr_ty, xr_cy_list, rg_curve, rg_params):
    return

    with plt.Dp():
        fig, ax = plt.subplots()
        axt = ax.twinx()
        axt.grid(False)

        ax.plot(x, y, color="orange")
        for cy in xr_cy_list[:-1]:
            ax.plot(x, cy, ":")

        ax.plot(x, ty, ":", color="red")

        # axt.plot(ratios, "-", color="C1")
        fig.tight_layout()
        plt.show()

def safe_ratios(ones, cy, ty, debug=False):
    ratios = cy/ty
    ratios[ty==0] = 1

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("safe_ratios entry")
            axt = ax.twinx()
            axt.grid(False)

            ax.plot(ty, ":")
            ax.plot(cy, ":")
            axt.plot(ratios, "-", color="C1")
            fig.tight_layout()
            plt.show()

    outliers = np.where(np.abs(ratios) > ones*OUTLIER_SCALE)[0]
    if len(outliers) == 0:
        return ratios

    # outliers appear only where both cy and ty are very small
    outlier_height = np.max(np.abs(ty[outliers]))
    if outlier_height < OUTLIER_IGNORE_LIMIT:
        widened = ty < OUTLIER_IGNORE_LIMIT
    else:
        widened = None

    if debug:
        print("outlier_height=", outlier_height)
        ratios_orig = ratios.copy()
        if widened is None:
            ratios[outliers] = 1
        else:
            ratios[widened] = 1

        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("safe_ratios outliers")

            axt = ax.twinx()
            axt.grid(False)

            ax.plot(cy)
            ax.plot(ty)

            axt.plot(ratios_orig)
            axt.plot(ratios)
            plt.show()
    else:
        if widened is None:
            ratios[outliers] = 1
        else:
            ratios[widened] = 1

    return ratios



"""
    ScatteringViewUtils.py

    Copyright (c) 2018-2023, SAXS Team, KEK-PF
"""
import numpy                as np

VERY_SMALL_ANGLE_LIMIT  = 0.03
NUM_SAMPLE_CURVES       = 20
ZLIM_EXPAND             = 3.0

def is_to_plot( index, i, q, very_small_angle_only=False ):
    # return q >= 0.01 and ( i == index or i % 8 == 0 )
    if very_small_angle_only:
        return q < VERY_SMALL_ANGLE_LIMIT or i == index
    else:
        return i == index or i % 8 == 0

def compute_baselines( qvector, index, corrector ):

    vsa_base_list = []
    all_base_list = []
    for i, q  in enumerate( qvector ):
        i_select = ( i == index or i % 8 == 0 )
        if q < VERY_SMALL_ANGLE_LIMIT or i_select:
            baseline = corrector.correct_a_single_q_plane( i, return_baseline=True )
            if q < VERY_SMALL_ANGLE_LIMIT or i == index:
                vsa_base_list.append( baseline )
            if i_select:
                all_base_list.append( baseline )

    return vsa_base_list, all_base_list

def draw_3d_scattering( ax, data, qvector, index, xray_curve_y,
                        title,
                        vsa_base_list, all_base_list,
                        zlim=None, zlim_expand=None, experimental=False,
                        very_small_angle_only=False,
                        sim_base=None ):

    ax.set_title( title, y=1.1 )
    # ax.set_xlabel( '\nQ(Å⁻¹)' )
    ax.set_xlabel( '\nQ($Å^{-1}$)' )
    ax.set_ylabel( '\nElution №' )
    ax.set_zlabel( '\nIntensity' )
    # ax.zaxis._set_scale('log')

    if zlim is None:
        if zlim_expand is None:
            zlim_expand = ZLIM_EXPAND
        zmin    = np.min( xray_curve_y )
        zmax    = np.max( xray_curve_y )
        # zmin, zmax = ax.get_zlim()
        zmin_   = zlim_expand * zmin + ( 1-zlim_expand ) * zmax
        zmax_   = ( 1-zlim_expand ) * zmin + zlim_expand * zmax
        zlim    = ( zmin_, zmax_ )

    ax.set_zlim( zlim )

    print( 'data.shape=', data.shape )

    size = data.shape[0]
    Y = np.arange( size )

    peak_j = None
    i_ = -1
    i_list = []
    for i, q in enumerate( qvector ):
        if not is_to_plot( index, i, q, very_small_angle_only=very_small_angle_only ):
            continue

        i_list.append( i )

        X = np.ones( size ) * q

        if i == index:
            alpha   = 1
            color   = 'orange'
            Z = xray_curve_y
        else:
            alpha   = 0.2
            color   = '#1f77b4'
            if len(data.shape) == 3:
                Z = data[:,i,1]
            else:
                Z = data[:,i]

        ax.plot( X, Y, Z, color=color, alpha=alpha )

        i_ += 1
        if very_small_angle_only:
            baseline = vsa_base_list[i_]
        else:
            baseline = all_base_list[i_]
        ax.plot( X, Y, baseline, color='red', alpha=0.2 )

    iv = np.array(i_list)

    if very_small_angle_only:
        line_array = np.array( vsa_base_list )
    else:
        line_array = np.array( all_base_list )

    Y_ = np.linspace( 0, size-1, NUM_SAMPLE_CURVES, dtype=int )

    for k, j in enumerate( Y_ ):
        x = qvector[iv]
        y = np.ones( len(iv) ) * j
        z = line_array[:,j]
        ax.plot( x, y, z, color='yellow', alpha=0.5 )
        if sim_base is not None:
            z = sim_base[j,iv]
            ax.plot( x, y, z, ':', color='cyan', alpha=0.5 )

    if experimental:
        ax.text2D( 0.15, 0.7, "Experimental Feature", transform=ax.transAxes, alpha=0.2, fontsize=50 )



"""
    EdPlotter.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.cm as cm
from matplotlib import animation
from .EdBoundary import guess_boundary_value

class EdPlotter:
    def __init__(self, fig, ax, data, cmap, file, denss_results=True, log_folder=None):
        self.pause = False
        self.fig = fig
        self.ax = ax
        self.total_angle = 360
        # fig.canvas.mpl_connect('button_press_event', self.on_click)

        n = data.shape[0]
        boundary = guess_boundary_value(data, debug=False)
        # print('boundary=', boundary)
        w = np.where(data > boundary)
        wi = np.array(w, dtype=int).T

        xyz = wi - n/2
        # print(xyz.shape)
        # print(xyz)

        cx, cy, cz = np.average(xyz, axis=0)
        x = xyz[:,0]
        y = xyz[:,1]
        z = xyz[:,2]
        d = (x-cx)**2 + (y-cy)**2 + (z-cz)**2
        m = np.average(d)
        s = np.std(d)
        s3y = m+s*3

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            plt.plot(d)
            plt.plot([0, len(d)-1], [m, m], ':')
            plt.plot([0, len(d)-1], [s3y, s3y], ':')
            plt.show()

        # remove too distant voxels
        xyz_ = xyz[d<s3y,:]
        w_ = (w[0][d<s3y], w[1][d<s3y], w[2][d<s3y])
        self.v = v = data[w_]

        vmin = v.min()
        vmax = v.max()
        vmin_ = vmin*1.1 + vmax*(-0.1)
        vmax_ = vmin*(-0.1) + vmax*1.1

        ax.set_axis_off()
        ax.set_title("Density Distribution in Voxels", fontsize=20, y=1.08)

        self.sc = sc = ax.scatter3D(xyz_[:,0], xyz_[:,1], xyz_[:,2], vmin=vmin_, vmax=vmax_, c=v, cmap=cmap, alpha=1, s=30)

        if denss_results:
            """
            Since the texts below must be opsitioned with figure coordinates,
            we need to trasform the box position.

            c.f.
            https://stackoverflow.com/questions/41267733/getting-the-coordinates-of-a-matplotlib-annotation-label-in-figure-coordinates
            """
            box = ax.patch.get_extents()
            tcbox = fig.transFigure.inverted().transform(box)
            x0, y0 = tcbox[0,:]
            x1, y1 = tcbox[1,:]
            # print('x0, x1=', x0, x1)
            tx1 = x0*0.99 + x1*0.01
            tx2 = x0*0.77 + x1*0.23
            tx3 = x0*0.6  + x1*0.4

            if log_folder is not None:
                try:
                    ret = get_log_items(log_folder + '/denss.log')
                    for xy, text in [ ((tx1, y0), r'$\chi^2$=%.3g' % ret.get('Chi2')),
                                    ((tx2, y0), r'$R_g$=%.3g' % ret.get('Rg')),
                                    ((tx3, y0), r'Support Volume=%.3g' % ret.get('Support Volume'))
                                    ]:
                        fig.text(*xy, text, fontsize=16)
                except:
                    from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                    etb = ExceptionTracebacker()
                    print(etb)

        fig.colorbar(sc, ax=ax)
        self.set_cubic_limits(n//4, ax)

    def make_anim(self, random=False):
        fig = self.fig
        ax = self.ax

        def init():
            # print('init:')
            return []

        self.azim = 0
        self.elev = 10
        def animate_(i):
            # print('animate', i)
            if random:
                if not self.pause:
                    self.azim += np.sign(np.random.rand()-0.5)*10
                    self.elev += np.sign(np.random.rand()-0.5)*10
            else:
                if not self.pause:
                    self.azim += 1
                    if self.azim == self.total_angle:
                        self.azim = 0
            ax.view_init(elev=self.elev, azim=self.azim)
            return []

        self.anim = animation.FuncAnimation(fig, animate_, init_func=init,
                                   frames=self.total_angle, interval=20, blit=True)

    def set_cubic_limits(self, min_size, ax):
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        zmin, zmax = ax.get_zlim()

        size = np.max([min_size, xmax-xmin, ymax-ymin, zmax-zmin])/2

        def get_limits(_min, _max):
            _mid = (_min+_max)/2
            return _mid - size, _mid + size

        ax.set_xlim(*get_limits(xmin, xmax))
        ax.set_ylim(*get_limits(ymin, ymax))
        ax.set_zlim(*get_limits(zmin, zmax))

    def on_click(self, event):
        self.pause ^= True

def ed_scatter(fig, axes, data, file):
    filename = get_name_for_title(file)
    fig.suptitle('Electron Density from ' + filename, fontsize=24)
    fig.tight_layout()
    fig.subplots_adjust(top=0.85)

    ax1, ax2 = axes

    # cmap = cm.bwr
    # cmap = cm.hot
    cmap = cm.plasma

    esc = EdPlotter(fig, ax1, data, cmap, file)

    v = esc.v
    h = np.histogram(v)
    # print('h=', h)

    colors = cmap(h[1])
    ax2.set_title("Volume Proportions in Voxels", fontsize=20)
    ax2.pie(h[0], colors=colors, startangle=90, counterclock=False, center=(0, 0.2), radius=0.6)

    ax2.set_ylim(-1, 1)

    pc = len(v)/np.prod(data.shape)*100

    ax2.text(0.0, -0.8,
        "This chart shows the volume proportions within\n"
        "the visible marked voxels in the left 3D figure,\n"
        "which in all occupy only %.2g%% voxels of\n"
        "the whole %dx%dx%d cubic space." % (pc, *data.shape),
        ha='center', fontsize=16)

    return esc

def get_log_items(path):
    from molass_legacy.DENSS.DenssUtils import get_denss_log_items
    return get_denss_log_items(path)

def get_name_for_title(file):
    nodes = file.replace('\\', '/').split('/')
    name = '/'.join(nodes[-4:])
    return name



"""
    AutoExecManager.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import sys
import os
import getopt
import re

# execute this file with python this-path demo
if len(sys.argv) > 1 and sys.argv[1].find("demo") >= 0:
    this_dir = os.path.dirname(os.path.abspath( __file__ ))
    sys.path.append(this_dir + '/..')

import psutil
from tksheet import Sheet
from tkinter import messagebox
from molass_legacy.KekLib.OurTkinter import Tk, Dialog

def shortened_list(L):
    ret_list = []
    for item in L:
        if item.find("\\") >= 0 or item.find("/") >= 0:
            nodes = re.split(r"[\\/]", item)
            if len(nodes) > 3:
                item = "/".join([nodes[0], "...",nodes[-1]])
        ret_list.append(item)
    return ret_list

CMDLINE_COL = 4

class AutoExecManager(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "AutoExecManager", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        column_width = 90
        cmdline_column_width = 1200
        colnames = ["pid", "ppid", "proc name", "func code", "cmdline"]
        num_columns = len(colnames)
        data_list = [colnames]
        pid_dict = {}
        ppid_dict = {}
        row = 1
        for proc in psutil.process_iter():
            if proc.name()[0:7] == "pythonw":
                pid = proc.pid
                ppid = proc.ppid()
                pid_dict[pid] = (row, pid, ppid, proc)
                ppid_dict[ppid] = (row, pid, ppid, proc)
                cmdline = proc.cmdline()
                func_code = self.get_func_code(cmdline)
                data_list.append((pid, ppid, proc.name(), func_code, shortened_list(cmdline)))
                row += 1

        num_rows = row

        height = int(22*num_rows) + 60
        width = column_width*num_columns + 60 + cmdline_column_width - column_width

        sheet = Sheet(body_frame, data=data_list, column_width=column_width, width=width, height=height)
        sheet.column_width(column=CMDLINE_COL, width=cmdline_column_width)
        sheet.enable_bindings("single_select", "row_select", "right_click_popup_menu")
        sheet.popup_menu_add_command("Kill", self.kill_paired_processes)
        sheet.pack()
        self.data_list = data_list
        self.sheet = sheet
        self.pid_dict = pid_dict
        self.ppid_dict = ppid_dict

    def get_func_code(self, cmdline):
        if str(cmdline).find("--") >= 0:
            return ""
        optlist, args = getopt.getopt(cmdline[2:], 'c:w:f:n:i:d:m:s:r:t:p:T:M:')
        optdict = dict(optlist)
        func_code = optdict.get("-c", "")
        return func_code

    def kill_paired_processes(self):
        selected = self.sheet.get_currently_selected()
        print("kill_paired_processes", selected)
        row = selected.row
        row_data = self.data_list[row]
        print(row_data)
        cmdline = row_data[CMDLINE_COL]
        if cmdline[1].find("molass.py") >= 0:
            row_type = "monitor"
            monitor_pid = row_data[0]
            try:
                row_, optimizer_pid = self.ppid_dict[monitor_pid][0:2]
                row_data_ = self.data_list[row_]
                cmdline_ = row_data_[CMDLINE_COL]
                func_code = self.get_func_code(cmdline_)
            except:
                optimizer_pid = None
                func_code = ""
            print(row_type, monitor_pid, optimizer_pid, func_code)
        else:
            row_type = "optimizer"
            monitor_pid = row_data[1]
            optimizer_pid = row_data[0]
            func_code = self.get_func_code(cmdline)
            print(row_type, monitor_pid, optimizer_pid, func_code)

        monitor_proc = self.pid_dict[monitor_pid][3]
        if optimizer_pid is None:
            optimizer_proc = None
            cwd = ""
            message =  ("%s row is selected.\n"
                        "are you sure to kill the following monitor processe?\n"
                        "    monitor: %d\n"
                        %  (row_type, monitor_pid) )
        else:
            optimizer_proc = self.pid_dict[optimizer_pid][3]
            cwd = optimizer_proc.cwd()
            message =  ("%s row is selected.\n"
                        "are you sure to kill the following paired processes?\n"
                        "    monitor: %d\n"
                        "    optimizer: %d\n"
                        "    func_code: %s\n"
                        "    cwd: %s"
                        %  (row_type, monitor_pid, optimizer_pid, func_code, cwd) )
        ok = messagebox.askokcancel(title="Kill confirmation", message=message)
        if ok:
            if optimizer_proc is None:
                monitor_proc.kill()
                messagebox.showinfo(title="Killed notification", message="the monitor process has been killed.")
            else:
                for proc in [monitor_proc, optimizer_proc]:
                    proc.kill()
                messagebox.showinfo(title="Killed notification", message="the paired processes have been killed.")

def run_manager():
    from molass_legacy.KekLib.TkUtils import get_tk_root
    root = get_tk_root()

    def show_manager():
        ts = AutoExecManager(root)
        ts.show()
        root.quit()

    root.after(0, show_manager)
    root.mainloop()
    root.destroy()

if __name__ == '__main__':
    run_manager()



"""
    QuickAnalysis.RgDiffRatios.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data

def compute_rgr_legacy_impl(qv, M, E, P1, A1, A2):
    # called from Conc.ConcDepend

    Minv = np.linalg.pinv(M)
    W = np.dot(Minv, P1)
    Pe  = np.sqrt(np.dot(E**2, W**2))
    Ae = Pe[:,0]

    sg1 = SimpleGuinier(np.array([qv, A1, Ae]).T)
    sg2 = SimpleGuinier(np.array([qv, A2, Ae]).T)
    Rg1 = sg1.Rg
    Rg2 = sg2.Rg

    if Rg1 is None or Rg1 == 0 or Rg2 is None or Rg2 == 0:
        rdr = 0
    else:
        rdr = (Rg1 - Rg2)*2/(Rg1 + Rg2)

    return rdr

RDR12_COMPUTABLE_RDRLR_LIMIT = 0.01     # < 

class RgDiffRatios:
    def __init__(self, sd, x_curve):
        self.logger = logging.getLogger(__name__)
        self.sd = sd
        self.paired_ranges = x_curve.get_default_editor_ranges()

    def compute_rg_quartets(self, debug=False):
        D, E, qv, xr_curve = self.sd.get_xr_data_separate_ly()

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Elution.CurveUtils import simple_plot
            print("paired_ranges=", self.paired_ranges)
            with plt.Dp():
                fig, ax = plt.subplots()
                simple_plot(ax, xr_curve)
                fig.tight_layout()
                plt.show()

        x = xr_curve.x
        y = xr_curve.y
        pair_list = []
        for k, erec in enumerate(self.paired_ranges):
            print([k], erec)
            rg_pairs = []
            for f, t in erec:
                eslice = slice(f,t+1)
                D_ = D[:,eslice]
                E_ = E[:,eslice]
                c = y[eslice]
                c = c/np.max(c)
                M1 = get_denoised_data(D_, rank=1)
                M2 = get_denoised_data(D_, rank=2)
                C1 = np.array([c])
                C2 = np.array([c, c**2])
                C1inv = np.linalg.pinv(C1)
                C2inv = np.linalg.pinv(C2)
                P1 = M1 @ C1inv
                P2 = M2 @ C2inv

                Minv = np.linalg.pinv(D_)
                W = np.dot(Minv, P1)
                Pe  = np.sqrt(E_**2 @ W**2)
                Ae = Pe[:,0]

                A1 = P1[:,0]
                A2 = P2[:,0]

                sg1 = SimpleGuinier(np.array([qv, A1, Ae]).T)
                sg2 = SimpleGuinier(np.array([qv, A2, Ae]).T)
                Rg1 = sg1.Rg
                Rg2 = sg2.Rg
                if Rg1 is None:
                    Rg1 = np.nan
                if Rg2 is None:
                    Rg2 = np.nan
                rg_pairs.append((Rg1,Rg2))
            pair_list.append(np.array(rg_pairs).T)

        return pair_list

    def get_rank_hints(self):
        hints = []
        for k, quartet in enumerate(self.compute_rg_quartets()):
            if len(quartet[0,:]) == 1:
                # as in 20200630_5
                self.logger.info("%dth rdr is not computable due to a single range %s", k, str(self.paired_ranges))
                computable = False
                rdr12 = np.nan
            else:
                rgL, rgR = quartet[0,:]
                rdrLR = (rgL - rgR)*2/(rgL + rgR)
                print([k], quartet, rdrLR)
                computable = abs(rdrLR) < RDR12_COMPUTABLE_RDRLR_LIMIT
                if computable:
                    rg1 = np.average(quartet[0,:])
                    rg2 = np.average(quartet[1,:])
                    rdr12 = (rg1 - rg2)*2/(rg1 + rg2)
                else:
                    self.logger.info("%dth rdr is not computable due to rdrLR=%.3g >= %.3g", k, rdrLR, RDR12_COMPUTABLE_RDRLR_LIMIT)
                    rdr12 = np.nan
            hints.append((computable, rdr12))
        return hints

def spike(caller):
    from time import time

    t0 = time()

    sd = caller.judge_holder.sd
    paired_ranges = caller.mapper.x_curve.get_default_editor_ranges()

    rdrs = RgDiffRatios(sd, paired_ranges)
    hints = rdrs.get_rank_hints()
    print("hints=", hints)

    print("it took ", time() - t0)



"""
    from
        https://github.com/ando-lab/regals

        REGALS: a general method to deconvolve X-ray scattering data from evolving mixtures
        https://www.biorxiv.org/content/10.1101/2020.12.06.413997v1
"""
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from copy import deepcopy
import numpy as np
from scipy import sparse as sp
from scipy.linalg import eig
from scipy.sparse.linalg import spsolve

class regals:
    
    def __init__(self, I, err):
        self.I = I
        self.err = err
    
    def fit_concentrations(self, mix):
        
        mix = deepcopy(mix)
        
        H = mix.H_concentration
        [AA, Ab] = mix.concentration_problem(self.I, self.err)
        
        u = spsolve(AA + H, Ab)
        
        u = np.split(u, np.cumsum(mix.k_concentration)[:-1])
        
        mix.u_concentration = u
        return mix
    
    def fit_profiles(self, mix):
        
        mix = deepcopy(mix)
        
        H = mix.H_profile
        [AA, Ab] = mix.profile_problem(self.I, self.err)
        
        u = spsolve(AA + H, Ab)
        
        u = np.split(u, np.cumsum(mix.k_profile)[:-1])
        
        n = mix.norm_profile
        u = [uj / nj if nj != 0 else uj for uj, nj in zip(u, n)]
        
        mix.u_profile = u
        return mix
    
    def step(self, mix):
        
        new_mix = self.fit_concentrations(self.fit_profiles(mix));
        
        resid = (self.I - new_mix.I_reg) / self.err
        
        params = {}
        params['x2'] = np.mean(resid ** 2)
        params['delta_concentration'] = np.sum(np.abs(new_mix.concentrations - mix.concentrations),0)
        params['delta_profile'] = np.sum(np.abs(new_mix.profiles - mix.profiles),0)
        params['delta_u_concentration'] = np.array([np.sum(np.abs(nupk - upk)) for nupk, upk in zip(new_mix.u_concentration, mix.u_concentration)])
        params['delta_u_profile'] = np.array([np.sum(np.abs(nupr - upr)) for nupr, upr in zip(new_mix.u_profile, mix.u_profile)])
        
        return [new_mix, params, resid]
    
    def run(self, mix, stop_fun = None, update_fun = None):
        
        if stop_fun == None:
            stop_fun = lambda num_iter, params: [num_iter >= 10, 'max_iter']
        
        if update_fun == None:
            update_fun = lambda num_iter, new_mix, params, resid: True
        
        num_iter = 0
        while True:
            num_iter += 1
            [mix, params, resid] = self.step(mix)
            
            update_fun(num_iter, mix, params, resid)
            
            [if_exit, exit_cond] = stop_fun(num_iter, params)
            if if_exit:
                break
        
        return [mix, params, resid, exit_cond]



class mixture:
    
    def __init__(self, components, lambda_concentration = np.array([]), lambda_profile = np.array([]), u_concentration = [], u_profile = []):
        self.components = components
        self.lambda_concentration = lambda_concentration
        self.lambda_profile = lambda_profile
        self.u_concentration = u_concentration
        self.u_profile = u_profile
        
        if len(self.u_concentration) == 0:
            self.u_concentration = [comp.concentration.u0 for comp in components]
        
        if len(self.u_profile) == 0:
            self.u_profile = [comp.profile.u0 for comp in components]
        
        if len(self.lambda_concentration) == 0:
            self.lambda_concentration = np.zeros(self.Nc)
        
        if len(self.lambda_profile) == 0:
            self.lambda_profile = np.zeros(self.Nc)
    
    @property
    def Nc(self):
        return len(self.components)
    
    @property
    def Nx(self):
        return self.components[0].concentration.Nx
    
    @property
    def Nq(self):
        return self.components[0].profile.Nq
    
    @property
    def k_concentration(self):
        return np.array([comp.concentration.k for comp in self.components])
    
    @property
    def k_profile(self):
        return np.array([comp.profile.k for comp in self.components])
    
    @property
    def I_reg(self):
        return self.profiles @ self.concentrations.T
    
    def estimate_concentration_lambda(self,err,ng):
        AA = self.concentration_problem(np.zeros((self.Nq,self.Nx)),err,False).todense()
        
        split_pos = np.cumsum(self.k_concentration)[:-1]
        AA = [np.hsplit(AAi, split_pos) for AAi in np.vsplit(AA, split_pos)]
        
        ll = np.zeros(self.Nc)
        L = [comp.concentration.L for comp in self.components]
        for k in range(self.Nc):
            d = np.real(eig(AA[k][k], (L[k].T @ L[k]).todense())[0])
            ll[k] = mixture.ng2lambda(d, ng[k])
        
        return ll
    
    def estimate_profile_lambda(self,err,ng):
        AA = self.profile_problem(np.zeros((self.Nq,self.Nx)),err,False).todense()
        
        split_pos = np.cumsum(self.k_profile)[:-1]
        AA = [np.hsplit(AAi, split_pos) for AAi in np.vsplit(AA, split_pos)]
        
        ll = np.zeros(self.Nc)
        L = [comp.profile.L for comp in self.components]
        for k in range(self.Nc):
            d = np.real(eig(AA[k][k], (L[k].T @ L[k]).todense())[0])
            ll[k] = mixture.ng2lambda(d, ng[k])
        
        return ll
    
    def concentration_problem(self, I, err, calc_Ab = True):
        
        w = 1 / np.mean(err,1)
        
        A = [comp.concentration.A for comp in self.components]
        
        D = w[:,np.newaxis] * I
        y = self.profiles
        y = w[:,np.newaxis] * y
        
        AA = [[(y[:,k1] @ y[:,k2]) * (A[k1].T @ A[k2]) for k2 in range(self.Nc)] for k1 in range(self.Nc)]
        AA = sp.vstack((sp.hstack(tuple(AAi)) for AAi in AA))
        
        if calc_Ab == True:
            Ab = [A[k].T @ (D.T @ y[:,k]) for k in range(self.Nc)]
            Ab = np.hstack(tuple(Ab))
            return [AA, Ab]
        else:
            return AA
    
    def profile_problem(self, I, err, calc_Ab = True):
        
        w = 1 / np.mean(err,1)
        
        A = [comp.profile.A for comp in self.components]
        A = [sp.diags(w,0) @ Ai for Ai in A]
        
        D = w[:,np.newaxis] * I
        c = self.concentrations
        
        AA = [[sp.csr_matrix((c[:,k1] @ c[:,k2]) * (A[k1].T @ A[k2])) for k2 in range(self.Nc)] for k1 in range(self.Nc)]
        AA = sp.vstack((sp.hstack(tuple(AAi)) for AAi in AA))
        
        if calc_Ab == True:
            Ab = [A[k].T @ (D @ c[:,k]) for k in range(self.Nc)]
            Ab = np.hstack(tuple(Ab))
            return [AA, Ab]
        else:
            return AA
        
    def extract_concentration(self,I,err,k):
        
        notk = np.setdiff1d(np.arange(self.Nc), k)
        c = self.concentrations
        y = self.profiles
        D = I - y[:,notk] @ c[:,notk].T
        yk = y[:,k]
        w = 1/np.mean(err,1)
        m = (w**2 * yk) / (yk @ (w**2 * yk))
        pk = D.T @ m
        sigmak = np.sqrt(err.T**2 @ m**2)
        
        return [pk, sigmak]
    
    def extract_profile(self,I,err,k):
        
        notk = np.setdiff1d(np.arange(self.Nc), k)
        c = self.concentrations
        y = self.profiles
        D = I - y[:,notk] @ c[:,notk].T
        ck = c[:,k]
        m = ck/(ck @ ck)
        Ik = D @ m
        sigmak = np.sqrt(err**2 @ m**2)
        
        return [Ik, sigmak]

    @property
    def H_concentration(self):
        L = [comp.concentration.L for comp in self.components]
        B = [Lpk * (lbdpk ** 0.5) for Lpk, lbdpk in zip(L, self.lambda_concentration)]
        B = sp.block_diag(B)
        return B.T @ B
    
    @property
    def H_profile(self):
        L = [comp.profile.L for comp in self.components]
        B = [Lpr * (lbdpr ** 0.5) for Lpr, lbdpr in zip(L, self.lambda_profile)]
        B = sp.block_diag(B)
        return B.T @ B
    
    @property
    def concentrations(self):
        return np.hstack(tuple(comp.concentration.A @ upk[:,np.newaxis] for comp, upk in zip(self.components, self.u_concentration)))
    
    @property
    def profiles(self):
        return np.hstack(tuple(comp.profile.A @ upr[:,np.newaxis] for comp, upr in zip(self.components, self.u_profile)))
        
    @property
    def norm_concentration(self):
        return [comp.concentration.norm(upk) for comp, upk in zip(self.components, self.u_concentration)]
    
    @property
    def norm_profile(self):
        return [comp.profile.norm(upr) for comp, upr in zip(self.components, self.u_profile)]
    
    @staticmethod
    def ng2lambda(dd, ng):
        
        ng0 = np.count_nonzero(np.isinf(dd))
        
        dd = dd[np.logical_and(dd >= 0,~np.isinf(dd))]
        
        lambda_list = np.logspace(np.amax(np.log10(dd)) + 2, np.amin(np.log10(dd)) - 2, 51)
        
        ng_list = np.zeros(len(lambda_list))
        for j in range(len(ng_list)):
            ng_list[j] = ng0 + sum(dd / (dd + lambda_list[j]))
        
        if ng < ng_list[0]:
            optimal_lambda = np.inf
        elif ng > ng_list[-1]:
            optimal_lambda = 0
        else:
            optimal_lambda = 10 ** np.interp(ng, ng_list, np.log10(lambda_list))
        
        return optimal_lambda



class component:
    
    def __init__(self, concentration, profile):
        self.concentration = concentration
        self.profile = profile



class concentration_class:
    
    def __init__(self, concentration_type, *arg, **kwarg):
        self.concentration_type = concentration_type
        
        if self.concentration_type == 'simple':
            self._regularizer = concentration_simple(*arg, **kwarg)
        
        elif self.concentration_type == 'smooth':
            self._regularizer = concentration_smooth(*arg, **kwarg)
        
        else:
            raise ValueError('unexpected concentration type')
    
    def __getattr__(self,attr):
        return super().__getattribute__('_regularizer').__getattribute__(attr) #super() usage for deepcopy



class concentration_simple:
    
    def __init__(self, x, xmin, xmax):
        self.x = x
        self.xmin = xmin
        self.xmax = xmax
    
    @property
    def Nx(self):
        return len(self.x)
    
    @property
    def y0(self):
        return self.A @ self.u0
    
    @property
    def F(self):
        is_in_concentration = np.isin(self.x,self.w)
        ind_in_w = np.nonzero(self.x[:,np.newaxis] == self.w)[1] #assumes no repetition in x
        v = np.arange(self.Nx)
        return sp.csr_matrix((np.ones(self.Nw), (v[is_in_concentration], ind_in_w)), shape=(self.Nx, self.Nw))
    
    @property
    def k(self):
        return self.Nw
    
    @property
    def w(self):
        return self.x[np.logical_and(self.x >= self.xmin, self.x <= self.xmax)] #assumes no repetition in x
    
    @property
    def Nw(self):
        return len(self.w)
    
    @property
    def u0(self):
        return np.ones(self.k)
    
    @property
    def L(self):
        return sp.eye(self.Nw)
    
    @property
    def A(self):
        return self.F

    def norm(self,u):
        return np.mean(u**2)**0.5



class concentration_smooth:
    
    def __init__(self, x, xmin, xmax, Nw = 50, is_zero_at_xmin = True, is_zero_at_xmax = True):
        #Changed input parameter order because python requires default parameters to follow non-default ones
        self.x = x
        self.xmin = xmin
        self.xmax = xmax
        self.Nw = Nw
        self.is_zero_at_xmin = is_zero_at_xmin
        self.is_zero_at_xmax = is_zero_at_xmax
    
    @property
    def Nx(self):
        return len(self.x)
    
    @property
    def y0(self):
        return self.A @ self.u0
    
    @property
    def F(self):
        ix = np.searchsorted(self.w,self.x,side='right')
        ix = ix - 1
        is_in_concentration = np.logical_and(ix > -1, ix < self.Nw - 1)
        v = np.arange(self.Nx)
        ix = ix[is_in_concentration]
        v = v[is_in_concentration]
        u = (self.x[is_in_concentration] - self.w[ix]) * (1/self.dw)
        return sp.csr_matrix((np.concatenate((1-u,u)), (np.concatenate((v,v)), np.concatenate((ix,ix+1)))),shape = (self.Nx, self.Nw))
    
    @property
    def k(self):
        return self.Nw - self.is_zero_at_xmin - self.is_zero_at_xmax
    
    @property
    def dw(self):
        return (self.xmax - self.xmin)/(self.Nw-1)
    
    @property
    def w(self):
        return np.linspace(self.xmin, self.xmax, self.Nw)
    
    @property
    def u0(self):
        if self.is_zero_at_xmax and self.is_zero_at_xmin:
            u0_tmp =  1 - ( 2*(self.w-self.xmin)/(self.xmax-self.xmin) - 1) ** 2
            return u0_tmp[1:-1]
        elif self.is_zero_at_xmax and (not self.is_zero_at_xmin):
            u0_tmp = (self.xmax-self.w)/(self.xmax-self.xmin)
            return u0_tmp[:-1]
        elif (not self.is_zero_at_xmax) and self.is_zero_at_xmin:
            u0_tmp = (self.w-self.xmin)/(self.xmax-self.xmin)
            return u0_tmp[1:]
        else:
            return np.ones(self.Nw)
    
    @property
    def L(self):
        L_tmp = sp.csr_matrix((-0.5*np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(0,self.Nw-2))), shape=(self.Nw-2, self.Nw))+\
            sp.csr_matrix((np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(1,self.Nw-1))), shape=(self.Nw-2, self.Nw))+\
            sp.csr_matrix((-0.5*np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(2,self.Nw))), shape=(self.Nw-2, self.Nw))
        if self.is_zero_at_xmax:
            L_tmp = L_tmp[:,:-1]
        if self.is_zero_at_xmin:
            L_tmp = L_tmp[:,1:]
        return L_tmp
    
    @property
    def A(self):
        A_tmp = self.F
        if self.is_zero_at_xmax:
            A_tmp = A_tmp[:,:-1]
        if self.is_zero_at_xmin:
            A_tmp = A_tmp[:,1:]
        return A_tmp
    
    def norm(self,u):
        return np.mean(u**2)**0.5



class profile_class:
    
    def __init__(self, profile_type, *arg, **kwarg):
        self.profile_type = profile_type
        
        if self.profile_type == 'simple':
            self._regularizer = profile_simple(*arg, **kwarg)
        
        elif self.profile_type == 'smooth':
            self._regularizer = profile_smooth(*arg, **kwarg)
        
        elif self.profile_type == 'realspace':
            self._regularizer = profile_real_space(*arg, **kwarg)
        
        else:
            raise ValueError('unexpected profile type')
    
    def __getattr__(self,attr):
        return super().__getattribute__('_regularizer').__getattribute__(attr) #super() usage for deepcopy



class profile_simple:
    
    def __init__(self, q):
        self.q = q
    
    @property
    def Nq(self):
        return len(self.q)
    
    @property
    def y0(self):
        return self.A @ self.u0
    
    @property
    def F(self):
        return sp.eye(self.Nq)
    
    @property
    def k(self):
        return self.Nq

    @property
    def u0(self):
        return np.ones(self.k)
    
    @property
    def w(self):
        return np.ones(self.k)
    
    @property
    def Nw(self):
        return self.Nq
    
    @property
    def L(self):
        return sp.eye(self.Nq)
    
    @property
    def A(self):
        return self.F
        
    def norm(self,u):
        return np.mean(u**2)**0.5



class profile_smooth:
    
    def __init__(self, q, Nw = 50):
        self.q = q
        self.Nw = Nw
    
    @property
    def Nq(self):
        return len(self.q)
    
    @property
    def y0(self):
        return self.A @ self.u0
    
    @property
    def F(self):
        ix = np.searchsorted(self.w,self.q,side='right')
        ix = ix - 1
        ix[ix == -1] = 0
        ix[ix == self.Nw - 1] = self.Nw - 2
        v = np.arange(self.Nq)
        u = (self.q - self.w[ix]) * (1/self.dw)
        return sp.csr_matrix((np.concatenate((1-u,u)), (np.concatenate((v,v)), np.concatenate((ix,ix+1)))),shape = (self.Nq, self.Nw))
    
    @property
    def k(self):
        return self.Nw
    
    @property
    def u0(self):
        return np.ones(self.k)
    
    @property
    def qmin(self):
        return np.amin(self.q)
   
    @property
    def qmax(self):
        return np.amax(self.q)
    
    @property
    def dw(self):
        return (self.qmax - self.qmin) / (self.Nw - 1)
    
    @property
    def w(self):
        return np.linspace(self.qmin, self.qmax, self.Nw)
    
    @property
    def L(self):
        return sp.csr_matrix((-0.5*np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(0,self.Nw-2))), shape=(self.Nw-2, self.Nw))+\
            sp.csr_matrix((np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(1,self.Nw-1))), shape=(self.Nw-2, self.Nw))+\
            sp.csr_matrix((-0.5*np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(2,self.Nw))), shape=(self.Nw-2, self.Nw))
    
    @property
    def A(self):
        return self.F
    
    def norm(self,u):
        return np.mean(u**2)**0.5



class profile_real_space:
    
    def __init__(self, q, dmax, Nw = 50, is_zero_at_r0 = True, is_zero_at_dmax = True):
        self.q = q
        self.dmax = dmax
        self.Nw = Nw
        self.is_zero_at_r0 = is_zero_at_r0
        self.is_zero_at_dmax = is_zero_at_dmax
    
    @property
    def Nq(self):
        return len(self.q)
    
    @property
    def y0(self):
        return self.A @ self.u0
    
    @property
    def F(self):
        F_tmp = 4 * np.pi * self.dw * np.sinc(np.outer(self.q,self.w) / np.pi)
        F_tmp[:,[0,-1]] = 0.5*F_tmp[:,[0,-1]]
        return F_tmp
    
    @property
    def k(self):
        return self.Nw - int(self.is_zero_at_dmax) - int(self.is_zero_at_r0)
    
    @property
    def dw(self):
        return self.dmax / (self.Nw - 1)
    
    @property
    def w(self):
        return self.dmax * np.linspace(0,1,self.Nw)
    
    @property
    def u0(self):
        u0_tmp = 1 - (2 * self.w / self.dmax - 1) ** 2
        u0_tmp = u0_tmp / (4 * np.pi * self.dw * np.sum(u0_tmp))
        if self.is_zero_at_dmax:
            u0_tmp = u0_tmp[:-1]
        if self.is_zero_at_r0:
            u0_tmp = u0_tmp[1:]
        return u0_tmp
    
    @property
    def L(self):
        L_tmp = sp.csr_matrix((-0.5*np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(0,self.Nw-2))), shape=(self.Nw-2, self.Nw))+\
            sp.csr_matrix((np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(1,self.Nw-1))), shape=(self.Nw-2, self.Nw))+\
            sp.csr_matrix((-0.5*np.ones(self.Nw-2), (np.arange(0,self.Nw-2), np.arange(2,self.Nw))), shape=(self.Nw-2, self.Nw))
        if self.is_zero_at_dmax:
            L_tmp = L_tmp[:,:-1]
        if self.is_zero_at_r0:
            L_tmp = L_tmp[:,1:]
        return L_tmp
    
    @property
    def A(self):
        A_tmp = self.F
        if self.is_zero_at_dmax:
            A_tmp = A_tmp[:,:-1]
        if self.is_zero_at_r0:
            A_tmp = A_tmp[:,1:]
        return A_tmp
    
    def norm(self,u):
        weight = 4 * np.pi * self.dw * np.ones(self.Nw)
        weight[[0,-1]] = 0.5* weight[[0,-1]]
        if self.is_zero_at_dmax:
            weight = weight[:-1]
        if self.is_zero_at_r0:
            weight = weight[1:]
        return np.sum(weight * u)



"""

    ファイル名：   TkUtils.py

    処理内容：

    Copyright (c) 2016-2024, Masatsuyo Takahashi, KEK-PF

"""
import os
import sys
import re
from MultiMonitor import get_selected_monitor

DO_NO_ADJUST_GEOMETRY   = 'DO_NO_ADJUST_GEOMETRY'

def split_geometry( geometry ):
    n = geometry.split('+')
    s = n[0].split('x')
    w = int( s[0] )
    h = int( s[1] )
    x = 0 if len(n) < 2 else int( n[1] )
    y = 0 if len(n) < 3 else int( n[2] )
    return [ w, h, x, y ]

def join_geometry( w, h, x, y ):
    return '%dx%d+%d+%d' % ( w, h, x, y )

def geometry_fix( top, x, y ):
    W, H, X, Y = split_geometry( top.geometry() )
    top.geometry( join_geometry( W, H, x, y ) )
    top.update()

"""
    URL: http://effbot.org/tkinterbook/wm.htm
"""
def parsegeometry(geometry):
    m = re.match(r"(\d+)x(\d+)([-+]\d+)([-+]\d+)", geometry)
    if not m:
        raise ValueError("failed to parse geometry string")
    return map(int, m.groups())

do_not_adjust_geometry = os.environ.get( DO_NO_ADJUST_GEOMETRY )
"""
    TODO:
    some tests seem to go wrong with adjusted geometry.
    it may have been improved by self.udate() before applying adjusted geometry.
"""

def adjusted_geometry( geometry, monitor=None, width_margin=0, height_margin=0.15, loc=None, debug=False ):
    global max_monitor, monitors

    if do_not_adjust_geometry:
        return geometry

    if monitor is None:
        monitor = get_selected_monitor()

    try:
        w0, h0, x0, y0 = split_geometry( geometry )
        w1, h1, x1, y1 = monitor.width, monitor.height, monitor.x,  monitor.y
        if debug:
            print(w0, h0, x0, y0)
            print(w1, h1, x1, y1)

        w1_ = int(w1 * (1-width_margin) )
        h1_ = int( h1 * (1-height_margin) )

        if loc is None:
            pass
        elif loc == "center":
            x1 += w1//2 - w0//2
            y1 += h1//2 - h0//2
        else:
            assert False

        return join_geometry( min(w0, w1_), min(h0, h1_), x1+x0, y1+y0 )
    except:
        return geometry

def geometry_move( toplevel, parent, x=50, y=50 ):
    x_, y_ = (parent.winfo_rootx()+x, parent.winfo_rooty()+y)
    # print( 'geometry_move: x_, y_=',  x_, y_  )
    toplevel.geometry("+%d+%d" % ( x_, y_ ) )

def rational_geometry( self, parent, w_ratio=0.5, h_ratio=0.5 ):
    w, h, x, y = split_geometry( parent.geometry() )
    x_, y_ = (parent.winfo_rootx() + int(w*w_ratio), parent.winfo_rooty() + int(h*h_ratio))
    self.geometry("+%d+%d" % ( x_, y_ ) )

def is_low_resolution():
    try:
        monitor = get_selected_monitor()
        # print( monitor.width, monitor.height )
        return monitor.height < 800
    except:
        # it seems this can occur
        return False

def get_widget_geometry(widget):
    w = widget.winfo_width()
    h = widget.winfo_height()
    x = widget.winfo_rootx()
    y = widget.winfo_rooty()
    return (w, h, x, y)

def get_tk_root(loc=None, withdraw=True, debug=False):
    if debug:
        import logging
        logging.basicConfig(filename="get_tk_root-debug.log", level=logging.DEBUG)
        logger = logging.getLogger(__name__)
        logger.info("get_tk_root entry")
    import OurTk as Tk
    from molass_legacy.KekLib.DebugPlot import set_plot_env
    root = Tk.Tk()
    if debug:
        logger.info("root ok")
    adj_geometry = adjusted_geometry(root.geometry(), loc=loc)
    if debug:
        logger.info("adjusted_geometry ok")
    root.geometry(adj_geometry)
    if debug:
        logger.info("geometry ok")
    if withdraw:
        root.withdraw()
    if debug:
        logger.info("withdraw ok")
    root.update()
    if debug:
        logger.info("update ok")
    set_plot_env(root)
    if debug:
        logger.info("set_plot_env ok")
    return root

class PositionSynchronizer:
    # modifeid from https://stackoverflow.com/questions/45183914/tkinter-detecting-a-window-drag-event
    def __init__(self, leader, follower, debug=False):
        self.leader = leader
        self.follower = follower
        self.geometry_re = re.compile(r"^(.+)(\+\-?\d+\+\-?\d+)")
        self.drag_id = None
        self.debug = debug
        self.leader.bind('<Configure>', self.on_configure)

    def on_configure(self, event):
        # print("on_configure")

        if event.widget is self.leader: # do nothing if the event is triggered by one of root's children
            if self.drag_id is None:
                # action on drag start
                if self.debug:
                    print('start drag')
            else:
                # cancel scheduled call to stop_drag
                self.leader.after_cancel(self.drag_id)
                # print('dragging')
            # schedule stop_drag
            self.drag_id = self.leader.after(100, self.stop_drag)

    def stop_drag(self):
        if self.debug:
            print('stop drag')
        self.synchronize_position()
        self.drag_id = None

    def synchronize_position(self):
        g1 = self.leader.geometry()
        g2 = self.follower.geometry()
        """
        g1: WWWxHHH+AAA+BBB
        g2: ___x___+CCC+DDD
        g2: ___x___+AAA+BBB <=== new geometry
        """
        m = self.geometry_re.match(g1)
        if m:
            position = m.group(2)
            new_g = re.sub(self.geometry_re, lambda m_: m_.group(1) + position, g2)
            if self.debug:
                print("g1=", g1)
                print("g2=", g2)
                print("new_g=", new_g)
            self.follower.geometry(new_g)



"""
    SecTheory.LognormalPoreCF.py

    Copyright (c) 2023, SAXS Team, KEK-PF

    from
        2014a, Annamária Sepsey
        Molecular theory of size exclusion chromatography for wide pore size distributions

"""
import numpy as np

INV_R2PI = 1/np.sqrt(2*np.pi)

def lognormal_pore_phi_impl(w, rgv, Np, Tp, me, mp, mu, sigma, x0):
    m = np.exp(mu + sigma**2/2)
    s = np.sqrt((np.exp(sigma**2) - 1) * np.exp(2*mu + sigma**2))
    minrp = max(0, m - 2*s)
    maxrp = m + 5*s
